# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Custom exception and error handling logic.
"""

# Standard
from collections.abc import Iterable
from types import GeneratorType
from typing import Any
import os

# Local
from caikit.config import get_config

# dictionary mapping string log channel name to error handler instances
# there is only one error handler instance for each log channel name
_error_handlers = {}


def get(log_chan):
    """Get an error handler associated with a given alog log channel.  The same error handler will
    be returned if this function is called repeatedly with the same log channel.

    Args:
        name: alog channel
            An alog log channel.

    Returns:
        ErrorHandler: An instance of `ErrorHandler` associated with `log_chan`
            that can be used to perform various error checks and to raise
            exceptions while also automatically logging appropriate message on
            `log_chan`.
    """
    return _error_handlers.setdefault(log_chan.name, ErrorHandler(log_chan))


class ErrorHandler:
    """An error handler that provides reusable error checking methods and also handles logging
    error messages automatically.  Calling an error handler directly is equivalent to calling
    the `.log_raise` method.
    """

    def __init__(self, log_chan):
        """Create a new error handler that provides reusable error checking and automatic logging.

        Args:
            log_chan: alog channel
                The logging channel that this error handle will use for logging.
        """
        self.log_chan = log_chan

    def _handle_exception_messages(self, log_code, exception):
        """Handle number of exception log messages to avoid overflows"""
        # increment the log message counter attribute or add it if not present
        if hasattr(exception, "_caikit_core_nexception_log_messages"):
            exception._caikit_core_nexception_log_messages += 1
        else:
            exception._caikit_core_nexception_log_messages = 0

        caikit_config = get_config()

        # if less than max log messages, then omit a log message
        if (
            exception._caikit_core_nexception_log_messages
            < caikit_config.max_exception_log_messages
        ):
            self.log_chan.error(
                log_code, "exception raised: {}".format(repr(exception))
            )

        # if at the limit omit one message stating that we will no longer log
        elif (
            exception._caikit_core_nexception_log_messages
            == caikit_config.max_exception_log_messages
        ):
            self.log_chan.error(
                log_code,
                "reached MAX_EXCEPTION_LOG_MESSAGES of `{}`, will no log exception `{}`".format(
                    caikit_config.max_exception_log_messages, repr(exception)
                ),
            )

    def log_raise(self, log_code, exception, root_exception=None):
        """Log an exception with a log code and then re-raise it.  Using this instead of simply
        using the `raise` keyword with your exceptions will ensure that log message is emitted on
        the `error` level for the log channel associated with this handler.  This is invaluable for
        debugging in environments where stack traces are not available.

        Args:
            log_code (str): A log code with format `<COR12345678E>` where `COR`
                is a short code for the library (for `caikit_core` in this
                example) and where `12345678` is a unique eight-digit identifier
                (example generation in `scripts/cor_log_code`) and where `E` is
                an error level short-code, one of `{'fatal': 'F', 'error': 'E',
                'warning': 'W', 'info': 'I', 'trace': 'T', 'debug': 'D'}`.
            exception (Exception): A python exception object or an instance of
                any subclass of `Exception`.
            root_exception: Exception (Optional)
                A python exception object or an instance of any subclass of `Exception` which is
                considered a 'root' exception, wrapped by the preceding exception. Done in cases
                where we'd like to wrap a base exception with a custom exception, preserving the
                original's stack-trace
        Notes:
            The error handler will track the number log messages emitted for a given exception and
            stop logging after `MAX_EXCEPTION_LOG_MESSAGES` have been logged for a single instance
            of an exception.  This is to prevent pathological logging, e.g., repeatedly calling
            `log_raise` in a recursive function.
        """
        self._handle_exception_messages(log_code, exception)

        if root_exception:
            self._handle_exception_messages(log_code, root_exception)
            # reraise the exception chained with root_exception
            raise exception from root_exception

        # reraise the exception
        raise exception

    # calling an error handler is equivalent to calling the `.log_raise` method
    __call__ = log_raise

    def type_check(self, log_code, *types, allow_none=False, **variables):
        """Check for acceptable types for a given object.  If the type check fails, a log message
        will be emitted at the error level on the log channel associated with this handler and a
        `TypeError` exception will be raised with an appropriate message.  This check should be used
        to check for appropriate variable types.  For example, to verify that an argument passed to
        a function that expects a string is actually an instance of a string.

        Args:
            log_code (str): A log code with format `<COR90063501E>` where `COR`
                is a short code for the library (for `caikit_core` in this
                example) and where `90063501` is a unique eight-digit identifier
                (example generation in `scripts/cor_log_code`) and where `E` is
                an error level short-code, one of `{'fatal': 'F', 'error': 'E',
                'warning': 'W', 'info': 'I', 'trace': 'T', 'debug': 'D'}`.
            *types (type or None): Variadic arguments containing all acceptable
                types for `variables`.  If any values of `variable` are not any
                of `*types` then a log message will be emitted and a `TypeError`
                will be raised.  Multiple types may be specified as separate
                arguments. If no types are specified, then a `RuntimeError` will
                be raised.
            allow_none (bool): If `True` then the values of `variables` are
                allowed to take on the value of `None` without causing the type
                check to fail.  If `False` (default) then `None` values will
                cause the type check to fail.
            **variables (object): Variadic keyword arguments to be examined for
                acceptable type.  The name of the variable is used in log and
                error messages while its value is actually check against
                `types`.  Multiple keyword variables may be specified.  If no
                variables are specified, then a `RuntimeError` will be raised.

        Examples:
            # this will raise a `TypeError` because `foo` is not `None` or a `list` or `tuple`
            > error.type_check('<COR99962332E>', None, list, tuple, foo='hello world')

            # this type check verifies that `foo` and `bar` are both strings
            > error.type_check('<COR03761101E>', str, foo=foo, bar=bar)
        """
        if not get_config().enable_error_checks:
            return

        if not types:
            self(log_code, RuntimeError("invalid type check: no types specified"))

        if not variables:
            self(log_code, RuntimeError("invalid type check: no variables specified"))

        for name, variable in variables.items():
            if allow_none and variable is None:
                continue

            # check if variable is an instance of one of `types`
            if not isinstance(variable, types):
                type_name = type(variable).__name__
                valid_type_names = tuple(typ.__name__ for typ in types)
                if allow_none:
                    valid_type_names += (type(None).__name__,)

                # create, log and raise an appropriate exception
                self(
                    log_code,
                    TypeError(
                        "type check failed: variable `{}` has type `{}` "
                        "(fully qualified name `{}`) not in `{}`".format(
                            name, type_name, self._fqname(variable), valid_type_names
                        )
                    ),
                )

    def type_check_all(self, log_code, *types, allow_none=False, **variables):
        """This type check is similar to `.type_check` except that it verifies that each variable
        in `**variables` is either a `list` or a `tuple` and then checks that *all* of the items
        they contain are instances of a type in `*types`.  If `allow_none` is set to `True`, then
        the variable is allowed to be `None`, but the items in the `list` or `tuple` are not.

        Examples:
            # this type check will verify that foo is a `list` or `tuple` containing only `int`s
            > foo = (1, 2, 3)
            > error.type_check('<COR50993928E>', int, foo='hello world')

            # this type check allows `foo` to be `None`
            > error.type_check('<COR79540602E>', None, foo=None)

            # this type check fails because `foo` contains `None`
            > error.type_check('<COR87797257E>', None, int, foo=(1, 2, None, 3, 4))

            # this type check fails because `bar` contains a `str`
            # but not for any other reason
            > foo = [1, 2, 3]
            > bar = [4, 5, 'x']
            > baz = None
            > error.type_check('<COR40818868E>', None, int, foo=foo, bar=bar, baz=None)
        """
        if not get_config().enable_error_checks:
            return

        if not types:
            self(log_code, RuntimeError("invalid type check: no types specified"))

        if not variables:
            self(log_code, RuntimeError("invalid type check: no variables specified"))

        top_level_types = (Iterable,)
        invalid_types = (
            str,
            GeneratorType,
        )  # top level types that will fail the type check

        for name, variable in variables.items():
            if allow_none and variable is None:
                continue

            # log and raise if variable is not an Iterable
            if not isinstance(variable, top_level_types) or isinstance(
                variable, invalid_types
            ):
                type_name = type(variable).__name__
                valid_type_names = tuple(typ.__name__ for typ in top_level_types)
                if allow_none:
                    valid_type_names += (type(None).__name__,)

                self(
                    log_code,
                    TypeError(
                        "type check failed: variable `{}` has type `{}` not in `{}`".format(
                            name, type_name, valid_type_names
                        )
                    ),
                )

            # log and raise if any item is not in list of valid types
            for item in variable:
                if not isinstance(item, types):
                    type_name = type(item).__name__
                    valid_type_names = tuple(typ.__name__ for typ in types)

                    self(
                        log_code,
                        TypeError(
                            "type check failed: element of `{}` has type `{}` not in `{}`".format(
                                name, type_name, valid_type_names
                            )
                        ),
                    )

    def subclass_check(
        self, log_code: str, child_class: Any, *parent_classes, allow_none: bool = False
    ):
        """Check that the given child classes are valid types and that they
        derive from the given set of parent classes [issubclass(x, (y, z))]. If
        the subclass check fails, a log message will be emitted at the error
        level on the log channel associated with this handler and a `TypeError`
        exception will be raised with an appropriate message. This check should
        be used to check that a given class meets the interface of a parent
        class. For example, to verify that a class handle is a valid ModuleBase
        subclass.

        Args:
            log_code (str): A log code with format `<COR90063501E>` where `COR`
                is a short code for the library (for `caikit_core` in this
                example) and where `90063501` is a unique eight-digit identifier
                and where `E` is an error level short-code, one of `{'fatal':
                'F', 'error': 'E', 'warning': 'W', 'info': 'I', 'trace': 'T',
                'debug': 'D'}`.
            child_class (Any): The class to be examined for acceptable class
                inheritance.
            *parent_classes (type): Variadic arguments containing all acceptable
                parent types for `child_classes`.  If any values of
                `child_classes` are not a valid type derived from one of
                `*parent_classes` then a log message will be emitted and a
                `TypeError` will be raised. Multiple parent_classes may be
                specified as separate arguments. If no parent_classes are
                specified, then a `RuntimeError` will be raised.
            allow_none (bool): If `True` then the values of `child_classes` are
                allowed to take on the value of `None` without causing the
                subclass check to fail.  If `False` (default) then `None` values
                will cause the subclass check to fail.

        Examples:
            # this will raise a `TypeError` because `Foo` is not `None` or
            # derived from Bar
            > class Bar: pass
            > class Foo: pass
            > error.subclass_check('<COR99962332E>', Bar, Foo=Foo)

            # this type check verifies that `foo` and `bar` are both strings
            > error.type_check('<COR03761101E>', str, foo=foo, bar=bar)
        """
        if not get_config().enable_error_checks:
            return

        if allow_none and child_class is None:
            return

        if not parent_classes:
            self(
                log_code,
                RuntimeError("invalid subclass check: no parent_classes given"),
            )

        if not isinstance(child_class, type) or not issubclass(
            child_class, parent_classes
        ):
            self(
                log_code,
                TypeError(
                    "subclass check failed: {} is not a subclass of {}".format(
                        child_class,
                        parent_classes,
                    )
                ),
            )

    def value_check(self, log_code, condition, *args):
        """Check for acceptable values for a given object.  If this check fails, a log message will
        be emitted at the error level on the log channel associated with this handler and a
        `ValueError` exception will be raised with an appropriate message.  This check should be
        used for checking for appropriate values for variable instances.  For example, to check that
        a numerical value has an appropriate range.

        Args:
            log_code (str): A log code with format `<COR55705215E>` where `COR`
                is a short code for the library (for `caikit_core` in this
                example) and where `55705215` is a unique eight-digit identifier
                (example generation in `scripts/cor_log_code`) and where `E` is
                an error level short-code, one of `{'fatal': 'F', 'error': 'E',
                'warning': 'W', 'info': 'I', 'trace': 'T', 'debug': 'D'}`.
            condition (bool): A boolean value that should describe if this check
                passes `True` or fails `False`. Upon calling this function, this
                is typically provided as an expression, e.g., `0 < variable <
                1`.
            *args: A variable set of arguments describing the value check that failed. If no
                args are provided then an empty msg string is assumed and no additional
                information will be provided, otherwise the first argument will be treated as 'msg'
                argument. Note that string interpolation can be lazily performed on `msg` using `{}`
                format syntax by passing additional arguments.  This is the preferred method for
                performing string interpolation on `msg` so that it is only done if an error
                condition is encountered.

        """
        if not get_config().enable_error_checks:
            return

        if not condition:
            interpolated_msg = (
                ""
                if not args
                else (args[0] if len(args) == 1 else args[0].format(*args[1:]))
            )

            self(
                log_code, ValueError("value check failed: {}".format(interpolated_msg))
            )

    def file_check(self, log_code, *file_paths):
        """Check to see if one or more file paths exist and are regular files.  If any do not exist
        or are not files, then a log message will be emitted on the log channel associated with this
        error handler and a `FileNotFoundError` will be raised with an appropriate error message.

        Args:
            log_code (str): A log code with format `<COR73692990E>` where `COR`
                is a short code for the library (for `caikit_core` in this
                example) and where `55705215` is a unique eight-digit identifier
                (example generation in `scripts/cor_log_code`) and where `E` is
                an error level short-code, one of `{'fatal': 'F', 'error': 'E',
                'warning': 'W', 'info': 'I', 'trace': 'T', 'debug': 'D'}`.
            *file_paths (str): Variadic argument containing strings specifying
                the file paths to check.  If any of these file paths does not
                exist or is not a regular file, then a log message will be
                emitted and a `FileNotFoundError` will be raised.
        """
        if not get_config().enable_error_checks:
            return

        for file_path in file_paths:
            if not os.path.exists(file_path):
                self(
                    log_code,
                    FileNotFoundError(
                        "File path `{}` does not exist".format(file_path)
                    ),
                )

            if not os.path.isfile(file_path):
                self(
                    log_code,
                    FileNotFoundError("Path `{}` is not a file".format(file_path)),
                )

    def dir_check(self, log_code, *dir_paths):
        """Check to see if one or more directory paths exist and are, in fact, directories.  If any
        do not exist then a `FileNotFoundError` will be raised and if they are not directories then
        a `NotADirectoryError` will be raised.  In either case, a log message will be emitted on the
        log channel associated with this error handler.

        Args:
            log_code (str): A log code with format `<COR63462828E>` where `COR`
                is a short code for the library (for `caikit_core` in this
                example) and where `55705215` is a unique eight-digit identifier
                (example generation in `scripts/cor_log_code`) and where `E` is
                an error level short-code, one of `{'fatal': 'F', 'error': 'E',
                'warning': 'W', 'info': 'I', 'trace': 'T', 'debug': 'D'}`.
            *dir_paths (str): Variadic argument containing strings specifying
                the directory paths to check.  If any of these file paths does
                not exist or is not a regular file, then a log message will be
                emitted and a `FileNotFoundError` or `NotADirectoryError` will
                raised.
        """
        if not get_config().enable_error_checks:
            return

        for dir_path in dir_paths:
            if not os.path.exists(dir_path):
                self(
                    log_code,
                    FileNotFoundError(
                        "Directory path `{}` does not exist".format(dir_path)
                    ),
                )

            if not os.path.isdir(dir_path):
                self(
                    log_code,
                    NotADirectoryError("Path `{}` is not a directory".format(dir_path)),
                )

    def _fqname(self, o) -> str:
        try:
            class_ = o.__class__
            return ".".join([class_.__module__, class_.__qualname__])
        except Exception:  # pylint: disable=broad-exception-caught
            return str(type(o))



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Common serialization interfaces that are generally helpful for saving out models that
are not necessarily specific to any domain's 3rd party libraries.
"""
# Standard
import abc

# Local
from . import fileio


class ObjectSerializer(abc.ABC):
    """Abstract class for serializing an object to disk."""

    @abc.abstractmethod
    def serialize(self, obj, file_path):
        """Serialize the provided object to the specified file path.

        Args:
            obj (object): The object to serialize
            file_path (str): Absolute path to which the object should be
                serialized
        """


class JSONSerializer(ObjectSerializer):
    """An ObjectSerializer for serializing to a JSON file."""

    def serialize(self, obj, file_path):
        """Serialize the provided object to a JSON file.

        Args:
            obj (object): The object to serialize
            file_path (str): Absolute path to which the object should be
                serialized
        """
        fileio.save_json(obj, file_path)


class TextSerializer(ObjectSerializer):
    """An ObjectSerializer for serializing a python list to a text file."""

    def serialize(self, obj, file_path):
        """Serialize the provided python list to a text file.

        Args:
            obj (list(str)): The list to serialize
            file_path (str): Absolute path to which the object should be
                serialized
        """
        lines = "\n".join(obj)
        fileio.save_txt(lines, file_path)


class YAMLSerializer(ObjectSerializer):
    """An ObjectSerializer for serializing to a YAML file."""

    def serialize(self, obj, file_path):
        """Serialize the provided object to a YAML file.

        Args:
            obj (object): The object to serialize
            file_path (str): Absolute path to which the object should be
                serialized
        """
        fileio.save_yaml(obj, file_path)


class CSVSerializer(ObjectSerializer):
    """An ObjectSerializer for serializing to a CSV file."""

    def serialize(self, obj, file_path):
        """Serialize the provided object to a CSV file.

        Args:
            obj (object): The object to serialize
            file_path (str): Absolute path to which the object should be
                serialized
        """
        fileio.save_csv(obj, file_path)


class PickleSerializer(ObjectSerializer):
    """An ObjectSerializer for pickling arbitrary Python objects."""

    def serialize(self, obj, file_path):
        """Serialize the provided object to a CSV file.

        Args:
            obj (any): The object to serialize
            file_path (str): Absolute path to which the object should be
                serialized
        """
        fileio.save_pickle(obj, file_path)



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This file contains interfaces required for unions of lists
"""

# Standard
from typing import List

# First Party
from py_to_proto.dataclass_to_proto import Annotated, FieldNumber

# Local
from caikit.core.data_model import PACKAGE_COMMON, DataObjectBase, dataobject


@dataobject(PACKAGE_COMMON)
class IntSequence(DataObjectBase):
    values: Annotated[List[int], FieldNumber(1)]


@dataobject(PACKAGE_COMMON)
class FloatSequence(DataObjectBase):
    values: Annotated[List[float], FieldNumber(1)]


@dataobject(PACKAGE_COMMON)
class StrSequence(DataObjectBase):
    values: Annotated[List[str], FieldNumber(1)]


@dataobject(PACKAGE_COMMON)
class BoolSequence(DataObjectBase):
    values: Annotated[List[bool], FieldNumber(1)]



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Local
from caikit.interfaces.nlp.data_model import Token, TokenizationResults

## Setup #########################################################################

dummy_token = Token(start=0, end=11, text="Hello World")
dummy_token2 = Token(start=0, end=7, text="foo bar")
tokenization_result = TokenizationResults([dummy_token, dummy_token2])

## Tests ########################################################################

### Token


def test_token_all_fields_accessible():
    token = Token(start=0, end=11, text="Hello World")
    assert token.start == 0
    assert token.end == 11
    assert token.text == "Hello World"


def test_token_from_proto_and_back():
    new = Token.from_proto(dummy_token.to_proto())
    assert new.start == 0
    assert new.end == 11
    assert new.text == "Hello World"


def test_token_from_json_and_back():
    new = Token.from_json(dummy_token.to_json())
    assert new.start == 0
    assert new.end == 11
    assert new.text == "Hello World"


### TokenizationResults


def test_tokenization_result_all_fields_accessible():
    tokenization_result = TokenizationResults([dummy_token, dummy_token2])
    assert tokenization_result.results[0] == dummy_token
    assert tokenization_result.results[1] == dummy_token2


def test_tokenization_result_from_proto_and_back():
    new = TokenizationResults.from_proto(tokenization_result.to_proto())
    assert new.results[0] == dummy_token
    assert new.results[1] == dummy_token2


def test_tokenization_result_from_json_and_back():
    new = TokenizationResults.from_json(tokenization_result.to_json())
    assert new.results[0] == dummy_token
    assert new.results[1] == dummy_token2



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Core base class for Augmentor combination schemes.
"""
# Standard
import random

# First Party
import alog

# Local
from ...exceptions import error_handler
from .. import AugmentorBase

log = alog.use_channel("AUG_SCHEME_BASE")
error = error_handler.get(log)


class SchemeBase:
    def __init__(self, preserve_order, augmentors, random_seed):
        """Initialize the core components of a merging scheme to be leveraged when combining
        augmentors.

        Args:
            preserve_order (bool): Indicates whether or not the contained
                augmentors should always be considered in the order that they
                were provided when they are being applied.
            augmentors (list(AugmentorBase) | tuple(AugmentorBase)): List or
                tuple of Augmentor objects to be applied.
            random_seed (int): Random seed for controlling shuffling behavior.
        """
        error.type_check("<COR54555981E>", bool, preserve_order=preserve_order)
        error.type_check("<COR54155111E>", list, tuple, augmentors=augmentors)
        error.type_check("<COR73170110E>", int, random_seed=random_seed)
        error.value_check(
            "<COR67355718E>",
            len(augmentors) > 0,
            "Must provide at least one augmentor to build a scheme.",
        )
        error.type_check_all("<COR37249765E>", AugmentorBase, augmentors=augmentors)
        # Determine whether or not augmentors should be applied in the order provided or
        # applied in random order.
        self._preserve_order = preserve_order
        self._current_order = list(range(len(augmentors)))
        self._augmentors = augmentors
        self._init_state = random.getstate()

    def execute(self, obj):
        """Execute the merged scheme, i.e., augment the object by leveraging the encapsulated
        augmentors.

        Args:
            obj (str | caikit.core.data_model.DataBase): Object to be augmented.
        Returns:
            str | caikit.core.data_model.DataBase: Augmented object of same type
                as input obj.
        """
        if not self._preserve_order:
            random.shuffle(self._current_order)
        return self._execute(obj)

    def reset(self):
        """Reset the random state of all encapsulated augmentors and the scheme itself."""
        # Reset the random state for all augmentors
        for aug in self._augmentors:
            aug.reset()
        # Reset the random state for the scheme using the default random package
        random.setstate(self._init_state)



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This toolkit utility provides common factory construction semantics and a common
base class for classes that can be constructed via caikit config
"""

# Standard
from typing import Optional, Type
import abc

# First Party
import aconfig
import alog

# Local
from ..exceptions import error_handler

log = alog.use_channel("FCTRY")
error = error_handler.get(log)


class FactoryConstructible(abc.ABC):
    """A class can be constructed by a factory if its constructor takes exactly
    one argument that is an aconfig.Config object and it has a name to identify
    itself with the factory.
    """

    @property
    @classmethod
    @abc.abstractmethod
    def name(cls) -> str:
        """This is the name of this constructible type that will be used by
        the factory to identify this class
        """

    @abc.abstractmethod
    def __init__(self, config: aconfig.Config, instance_name: str):
        """A FactoryConstructible object must be constructed with a config
        object that it uses to pull in all configuration
        """


class Factory:
    """The base factory class implements all common factory functionality to
    read a designated portion of config and instantiate an instance of the
    registered classes.
    """

    # The keys in the instance config
    TYPE_KEY = "type"
    CONFIG_KEY = "config"

    def __init__(self, name: str):
        """Construct with the path in the global config where this factory's
        configuration lives.
        """
        self._name = name
        self._registered_types = {}

    @property
    def name(self) -> str:
        return self._name

    def register(self, constructible: Type[FactoryConstructible]):
        """Register the given constructible"""
        current = self._registered_types.get(constructible.name)
        error.value_check(
            "<COR27473932E>",
            current is None or current is constructible,
            "Conflicting registration of {}",
            constructible.name,
        )
        self._registered_types[constructible.name] = constructible

    def construct(
        self,
        instance_config: dict,
        instance_name: Optional[str] = None,
    ) -> FactoryConstructible:
        """Construct an instance of the given type"""
        inst_type = instance_config.get(self.__class__.TYPE_KEY)
        inst_cfg = aconfig.Config(
            instance_config.get(self.__class__.CONFIG_KEY, {}),
            override_env_vars=False,
        )
        inst_cls = self._registered_types.get(inst_type)
        error.value_check(
            "<COR41162423E>",
            inst_cls is not None,
            "No {} class registered for type {}",
            self.name,
            inst_type,
        )
        instance_name = instance_name or inst_cls.name
        return inst_cls(inst_cfg, instance_name)



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defines the backend for making Image data model objects work nicely with PIL
images. The ImagePilBackend outlines the manner in which we coerce our inputs to PIL
images, retrieve/cache data model attributes from the encapsulated PIL image, and so on.
"""

# Standard
from typing import Any, Union
import io
import os
import pathlib

# Third Party
from PIL import Image as PILImage
import numpy as np

# First Party
import alog

# Local
from caikit.core import error_handler
from caikit.core.data_model.base import DataBase
from caikit.core.data_model.data_backends import DataModelBackendBase

log = alog.use_channel("DATABACK")
error = error_handler.get(log)
PIL_SOURCE_TYPES = Union[PILImage.Image, pathlib.PosixPath, str, np.ndarray, bytes]


class ImagePilBackend(DataModelBackendBase):
    def __init__(self, image_data: PIL_SOURCE_TYPES):
        self._image_data = self.__class__.coerce_to_pil(image_data)
        # Unless the PIL image explicitly has information about its format, export
        # serialized image data as a PNG since it's not a lossy compression
        self._export_format = (
            self._image_data.format if self._image_data.format is not None else "PNG"
        )

    def get_attribute(self, data_model_class: DataBase, name: str) -> Any:
        # Convert the underlying PIL image to bytes
        if name == "image_data":
            # Save the whole contents of the image into a BytesIO object; this includes image fmt
            im_arr = io.BytesIO()
            self._image_data.save(im_arr, format=self._export_format)
            # Then return the whole value of the buffer as a bytes object
            return im_arr.getvalue()
        # Delegate to common parent logic
        return super().get_attribute(data_model_class, name)

    ##### Type coercion to PIL
    @classmethod
    def coerce_to_pil(
        cls,
        image_data: PIL_SOURCE_TYPES,
    ) -> PILImage:
        """Given an object representing an image in a wide variety of formats, force it into a
        PIL image representation.
        Supported formats:
            - [PIL.Image.Image] loaded PIL image; this is a no-op
            - [str] path to an image on disk to be loaded
            - [pathlib.PosixPath] path to an image on disk to be loaded
            - [np.ndarray] Numpy array of type uint8
            - [bytes] Image loaded into a bytes object

        Args:
            image_data:
                Raw data object to be coerced to a PIL image.
        Returns:
            PILImage.Image
                PIL image representation of the provided image data.
        """
        # If we're given a PIL image, then we have nothing to do!
        if isinstance(image_data, PILImage.Image):
            return image_data
        # Load a numpy array; [uint8]
        if isinstance(image_data, np.ndarray):
            return cls._coerce_from_numpy(image_data)
        # Load a path on disk (str or pathlib)
        if isinstance(
            image_data,
            (
                str,
                pathlib.PosixPath,
            ),
        ):
            return cls._coerce_from_path(image_data)
        # Load from a bytes object containing the whole image
        if isinstance(image_data, bytes):
            return cls._coerce_from_bytes(image_data)
        error(
            "<COR17413231E>",
            TypeError(
                f"Unsupported data type could not be coerced to PIL! {type(image_data)}"
            ),
        )

    @staticmethod
    def _coerce_from_numpy(image_data: np.ndarray) -> PILImage.Image:
        """Given image data as a Numpy array, load it as a PIL image.

        Args:
            image_data: np.ndarray
                Numpy array representing an image (uint8).
        Returns:
            PIL.Image.Image
                PIL image representation of the numpy array.
        """
        error.value_check(
            "<COR73134226E>",
            image_data.dtype == np.uint8,
            "Numpy array must be of type [np.uint8] to be loaded as a data model image",
        )
        return PILImage.fromarray(image_data)

    @classmethod
    def _coerce_from_path(cls, image_data: Union[str, pathlib.PosixPath]) -> PILImage:
        """Given a str, which we assume to be a path to an image on disk, or a Pathlib object,
        try to load it as a PIL image.

        Args:
            image_data: Union[pathlib.PosixPath, str]
                Path to be loaded.
        Returns:
            PIL.Image.Image
                PIL image representation of the image loaded from disk.
        """
        if isinstance(image_data, str) and not os.path.isfile(image_data):
            error(
                "<COR14433331E>",
                FileNotFoundError(f"Provided path [{image_data}] is not a file"),
            )
        elif isinstance(image_data, pathlib.PosixPath) and not image_data.is_file():
            error(
                "<COR14499831E>",
                FileNotFoundError(f"Provided pathlib [{image_data}] is not a file"),
            )
        return PILImage.open(image_data)

    @staticmethod
    def _coerce_from_bytes(image_data: bytes) -> PILImage:
        """Given bytes, which we assume to represent a full image, try to load it as a PIL image.
        Args:
            image_data: bytes
                binary data to be loaded as a PIL image.
        Returns:
            PIL.Image.Image
                PIL image representation of the bytes object.
        """
        return PILImage.open(io.BytesIO(image_data))

    ##### Backend views
    def as_numpy(self) -> np.ndarray:
        """Zero-copy method to produce the PIL image as a numpy array.
        Returns:
            np.ndarray
                Numpy array representation of the PIL image data.
        """
        return np.asarray(self._image_data)

    def as_pil(self) -> PILImage.Image:
        """Return a handle to our PIL image.
        Returns:
            PIL.Image.Image
                PIL image representation of this data model object.
        """
        return self._image_data



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Data structures for text representations"""

# Standard
from typing import List

# First Party
from py_to_proto.dataclass_to_proto import Annotated, FieldNumber
import alog

# Local
from ....core import DataObjectBase, dataobject
from .package import NLP_PACKAGE

log = alog.use_channel("DATAM")


@dataobject(package=NLP_PACKAGE)
class Token(DataObjectBase):
    """Tokens here are the basic units of text. Tokens can be characters, words,
    sub-words, or other segments of text or code, depending on the method of
    tokenization chosen or the task being implemented.
    """

    start: Annotated[int, FieldNumber(1)]  # Beginning offset of the token
    end: Annotated[int, FieldNumber(2)]  # Ending offset of the token
    text: Annotated[str, FieldNumber(3)]  # Text referenced by this token


@dataobject(package=NLP_PACKAGE)
class TokenizationResults(DataObjectBase):
    """Tokenization result generated from a text."""

    results: Annotated[List[Token], FieldNumber(1)]


@dataobject(package=NLP_PACKAGE)
class TokenizationStreamResult(TokenizationResults):
    """
    Streaming tokenization result that indicates up to where in stream is processed.
    """

    processed_index: Annotated[
        int, FieldNumber(2)
    ]  # Result index up to which text is processed



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Caikit Core AI Framework library.  This is the base framework for core AI/ML libraries.
"""

# the import order cannot adhere to the linter here because we must do things like
# disable warnings, initialize the JVM and configure logging in a specific order
# pylint: disable=wrong-import-position,wrong-import-order

# NOTE: There are cyclic imports due to the "import *"s here, when modules then
# "import core"

# Standard
# We're filtering (most) warnings for now
import warnings as _warnings

_warnings.filterwarnings("ignore")

# Local
from .data_model import DataObjectBase, dataobject
from .exceptions import DataValidationError, error_handler
from .model_manager import *
from .module_backends import *
from .modules import ModuleBase, ModuleConfig, ModuleLoader, ModuleSaver, module
from .task import TaskBase, task
from .toolkit import *

# Configure the global model wrangling functions
MODEL_MANAGER = ModelManager()
extract = MODEL_MANAGER.extract
load = MODEL_MANAGER.load
resolve_and_load = MODEL_MANAGER.resolve_and_load
train = MODEL_MANAGER.train
get_model_future = MODEL_MANAGER.get_model_future



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Standard
import os

# Third Party
# Import the script code
from dsconverter import CustomDocstringConverter
import pytest


# Test case for convert_to_google_style function
def test_get_docstrings():
    converter = CustomDocstringConverter()
    # Get the current directory
    current_directory = os.getcwd()

    # Specify the relative path to the file
    relative_path = "caikit/core/modules/config.py"

    # Join the current directory and the relative path to get the full file path
    test_path = os.path.join(current_directory, relative_path)
    docstrings = converter.extract_docstrings_from_file(test_path)
    assert len(docstrings) == 4


def test_basic_custom_docstring():
    converter = CustomDocstringConverter()
    test1 = '''
    """Method for extracting pred set from dataset. Implemented in subclass.

        Args:
            dataset:  object
                In-memory version of whatever
            preprocess_func:  method
                Function used as proxy for any preliminary steps that need to be taken to run the
                model on the input text. This helper function ultimately leads to the input to this
                module and may involve executing other modules.
            foo
                Description
            *args, **kwargs: dict
                Optional keyword arguments for prediction set extraction.
        Returns:
            list
                List of labels in the format of the module_type that is being called.
        """
    '''

    converted = converter.convert_to_google_style(test1)

    assert (
        converted
        == '''
    """Method for extracting pred set from dataset. Implemented in subclass.

        Args:
            dataset (object): In-memory version of whatever
            preprocess_func (method): Function used as proxy for any preliminary
                steps that need to be taken to run the model on the input text.
                This helper function ultimately leads to the input to this
                module and may involve executing other modules.
            foo
                Description
            *args, **kwargs (dict): Optional keyword arguments for prediction set extraction.
        Returns:
            list: List of labels in the format of the module_type that is being
                called.
       """
    '''
    )


# Test case for convert_to_google_style function
def test_complex_custom_docstring():
    converter = CustomDocstringConverter()
    test1 = '''
    """Time a model `load` call.

        Args:
            *args: list
                Will be passed to `self.load`.
            **kwargs:  dict
                Will be passed to `self.load` -- the only way to pass arbitrary arguments to
                `self.load` from this function.

        Returns:
            int, caikit.core._ModuleBase
                The first return value is the total time spent in the `self.load` call. The second
                return value is the loaded model.

        Notes:
            You can pass everything that should go to the run function normally using args/kwargs.
            Example: `model.timed_load("/model/path/dir")`
        """
    '''

    converted = converter.convert_to_google_style(test1)

    assert (
        converted
        == '''
    """Time a model `load` call.

        Args:
            *args (list): Will be passed to `self.load`.
            **kwargs (dict): Will be passed to `self.load` -- the only way to
                pass arbitrary arguments to `self.load` from this function.

        Returns:
            int, caikit.core._ModuleBase: The first return value is the total
                time spent in the `self.load` call. The second return value is
                the loaded model.

        Notes:
            You can pass everything that should go to the run function normally using args/kwargs.
            Example: `model.timed_load("/model/path/dir")`
        """
    '''
    )


def test_no_arg_desc_docstring():
    converter = CustomDocstringConverter()
    test1 = '''
    """Time a model `load` call.

        Args:
            *args: list
            **kwargs:  dict
            hello: foo

        Returns:
            int, caikit.core._ModuleBase
                The first return value is the total time spent in the `self.load` call. The second
                return value is the loaded model.

        Notes:
            You can pass everything that should go to the run function normally using args/kwargs.
            Example: `model.timed_load("/model/path/dir")`
        """
    '''

    converted = converter.convert_to_google_style(test1)

    assert (
        converted
        == '''
    """Time a model `load` call.

        Args:
            *args (list)
            **kwargs (dict)
            hello (foo)
        Returns:
            int, caikit.core._ModuleBase: The first return value is the total
                time spent in the `self.load` call. The second return value is
                the loaded model.

        Notes:
            You can pass everything that should go to the run function normally using args/kwargs.
            Example: `model.timed_load("/model/path/dir")`
        """
    '''
    )


def test_no_arg_description_returns():
    converter = CustomDocstringConverter()
    test1 = '''
    """Extract the core list of predictions that is needed for quality evaluation

        Args:
            pred_set: list
        Returns:
            pred_annotations: list
        """
    '''
    converted = converter.convert_to_google_style(test1)

    assert (
        converted
        == '''
    """Extract the core list of predictions that is needed for quality evaluation

        Args:
            pred_set (list)
        Returns:
            pred_annotations: list
        """
    '''
    )


def test_already_converted():
    converter = CustomDocstringConverter()
    test1 = '''
    """Time a model `load` call.

        Args:
             *args (list): Will be passed to `self.load`.
             **kwargs (dict): Will be passed to `self.load` -- the only way to
                pass arbitrary arguments to `self.load` from this function.

        Returns:
             int, caikit.core._ModuleBase: The first return value is the
               total time spent in the `self.load` call. The second return value is the
               loaded model.

        Notes:
            You can pass everything that should go to the run function normally using args/kwargs.
            Example: `model.timed_load("/model/path/dir")`
        """
        '''
    converted = converter.convert_to_google_style(test1)

    assert (
        converted
        == '''
    """Time a model `load` call.

        Args:
             *args (list): Will be passed to `self.load`.
             **kwargs (dict): Will be passed to `self.load` -- the only way to
                pass arbitrary arguments to `self.load` from this function.

        Returns:
             int, caikit.core._ModuleBase: The first return value is the
               total time spent in the `self.load` call. The second return value is the
               loaded model.

        Notes:
            You can pass everything that should go to the run function normally using args/kwargs.
            Example: `model.timed_load("/model/path/dir")`
        """
        '''
    )


def test_examples_in_docstring():
    converter = CustomDocstringConverter()
    test1 = '''
    """Create a new data stream from a python iterable, such as a list or tuple.  This data
        stream produces a single data item for each element of the iterable..

        Args:
            data:  iterable
                A list or tuple or other python iterable used to construct a new data stream where
                each data item contains a single data item.

        Returns:
            DataStream
                A new data stream that produces data items from the elements of `data`.

        Examples:
            >>> list_stream = DataStream.from_iterable([1, 2, 3])
            >>> for data_item in list_stream:
            >>>     print(data_item)
            1
            2
            3
        """
        '''
    converted = converter.convert_to_google_style(test1)

    assert (
        converted
        == '''
    """Create a new data stream from a python iterable, such as a list or tuple.  This data
        stream produces a single data item for each element of the iterable..

        Args:
            data (iterable): A list or tuple or other python iterable used to
                construct a new data stream where each data item contains a
                single data item.

        Returns:
            DataStream: A new data stream that produces data items from the
                elements of `data`.

        Examples:
            >>> list_stream = DataStream.from_iterable([1, 2, 3])
            >>> for data_item in list_stream:
            >>>     print(data_item)
            1
            2
            3
        """
        '''
    )


def test_no_data_type_arg():
    converter = CustomDocstringConverter()
    test1 = '''
    """Create a new data stream from a python iterable, such as a list or tuple.  This data
        stream produces a single data item for each element of the iterable..

        Args:
            proto:
                A protocol buffer to be populated.
        
        Returns:
            DataStream
                A new data stream that produces data items from the elements of `data`.

    """
    '''
    converted = converter.convert_to_google_style(test1)

    assert (
        converted
        == '''
    """Create a new data stream from a python iterable, such as a list or tuple.  This data
        stream produces a single data item for each element of the iterable..

        Args:
            proto: A protocol buffer to be populated.
        Returns:
            DataStream: A new data stream that produces data items from the
                elements of `data`.
       

    """
    '''
    )


def test_no_data_type_arg_same_line_desc():
    converter = CustomDocstringConverter()
    test1 = '''
    """Create a new data stream from a python iterable, such as a list or tuple.  This data
        stream produces a single data item for each element of the iterable..

        Args:
            proto: A protocol buffer to be populated.
        
        Returns:
            DataStream
                A new data stream that produces data items from the elements of `data`.

    """
    '''
    converted = converter.convert_to_google_style(test1)

    assert (
        converted
        == '''
    """Create a new data stream from a python iterable, such as a list or tuple.  This data
        stream produces a single data item for each element of the iterable..

        Args:
            proto: A protocol buffer to be populated.
        Returns:
            DataStream: A new data stream that produces data items from the
                elements of `data`.
       

    """
    '''
    )


def test_no_data_type_arg_multi_line_desc():
    converter = CustomDocstringConverter()
    test1 = '''
    """Create a new data stream from a python iterable, such as a list or tuple.  This data
        stream produces a single data item for each element of the iterable..

        Args:
            proto: 
                A protocol buffer to be populated.
                More description here
                More here!
        
        Returns:
            DataStream
                A new data stream that produces data items from the elements of `data`.

    """
    '''
    converted = converter.convert_to_google_style(test1)

    assert (
        converted
        == '''
    """Create a new data stream from a python iterable, such as a list or tuple.  This data
        stream produces a single data item for each element of the iterable..

        Args:
            proto: A protocol buffer to be populated.
                More description here
                More here!
        
        Returns:
            DataStream: A new data stream that produces data items from the
                elements of `data`.
       

    """
    '''
    )


def test_buncha_ors():
    converter = CustomDocstringConverter()
    test1 = '''
    """Create a new data stream from a python iterable, such as a list or tuple.  This data
        stream produces a single data item for each element of the iterable..

        Args:
            proto: int | int | int | int | int | int | int 
                A protocol buffer to be populated.
                More description here
                More here!
        
        Returns:
            DataStream
                A new data stream that produces data items from the elements of `data`.

    """
    '''
    converted = converter.convert_to_google_style(test1)

    assert (
        converted
        == '''
    """Create a new data stream from a python iterable, such as a list or tuple.  This data
        stream produces a single data item for each element of the iterable..

        Args:
            proto (int | int | int | int | int | int | int ): A protocol buffer
                to be populated. More description here More here!
        
        Returns:
            DataStream: A new data stream that produces data items from the
                elements of `data`.
       

    """
    '''
    )


def test_fewa_ors():
    converter = CustomDocstringConverter()
    test1 = '''
    """Create a new data stream from a python iterable, such as a list or tuple.  This data
        stream produces a single data item for each element of the iterable..

        Args:
            obj: str | caikit.core.data_model.DataBase
                Object to be augmented.
        Returns:
            str | caikit.core.data_model.DataBase
                Augmented object of same type as input obj.
        """
    '''
    converted = converter.convert_to_google_style(test1)

    assert (
        converted
        == '''
    """Create a new data stream from a python iterable, such as a list or tuple.  This data
        stream produces a single data item for each element of the iterable..

        Args:
            obj (str | caikit.core.data_model.DataBase): Object to be augmented.
        Returns:
            str | caikit.core.data_model.DataBase: Augmented object of same type
                as input obj.
       """
    '''
    )


def test_data_type_same_line_returns():
    converter = CustomDocstringConverter()
    test1 = '''
    """Configure caikit for your usage!
    Merges into the internal aconfig.Config object with overrides from multiple sources.
    Sources, last takes precedence:
        1. The existing configuration from calls to `caikit.configure()`
        2. The config from `config_yml_path`
        3. The config from `config_dict`
        4. The config files specified in the `config_files` configuration
            (NB: This may be set by the `CONFIG_FILES` environment variable)
        5. Environment variables, in ALL_CAPS_SNAKE_FORMAT
    Args:
        config_yml_path (Optional[str]): The path to the base configuration yaml
            with overrides for your usage.
        config_dict (Optional[Dict]): Config overrides in dictionary form
    Returns: None
        This only sets the config object that is returned by `caikit.get_config()`
    """
    '''
    converted = converter.convert_to_google_style(test1)

    assert (
        converted
        == '''
    """Configure caikit for your usage!
    Merges into the internal aconfig.Config object with overrides from multiple sources.
    Sources, last takes precedence:
        1. The existing configuration from calls to `caikit.configure()`
        2. The config from `config_yml_path`
        3. The config from `config_dict`
        4. The config files specified in the `config_files` configuration
            (NB: This may be set by the `CONFIG_FILES` environment variable)
        5. Environment variables, in ALL_CAPS_SNAKE_FORMAT
    Args:
        config_yml_path (Optional[str]): The path to the base configuration yaml
            with overrides for your usage.
        config_dict (Optional[Dict]): Config overrides in dictionary form
    Returns: None: This only sets the config object that is returned by
        `caikit.get_config()`
"""
    '''
    )


def test_example_usage():
    converter = CustomDocstringConverter()
    test1 = '''
    """The decorator for AI Task classes.

    This defines an output data model type for the task, and a minimal set of required inputs
    that all public models implementing this task must accept.

    As an example, the `caikit.interfaces.nlp.SentimentTask` might look like::

        @task(
            required_inputs={
                "raw_document": caikit.interfaces.nlp.RawDocument
            },
            output_type=caikit.interfaces.nlp.SentimentPrediction
        )
        class SentimentTask(caikit.TaskBase):
            pass

    and a public model that implements this task might have a .run function that looks like::

        def run(raw_document: caikit.interfaces.nlp.RawDocument,
                inference_mode: str = "fast",
                device: caikit.interfaces.common.HardwareEnum) ->
                    caikit.interfaces.nlp.SentimentPrediction:
            # impl

    Note the run function may include other arguments beyond the minimal required inputs for
    the task.

    Args:
        required_parameters (Dict[str, ValidInputTypes]): The required parameters that all public
            models' .run functions must contain. A dictionary of parameter name to parameter
            type, where the types can be in the set of:
                - Python primitives
                - Caikit data models
                - Iterable containers of the above
                - Caikit model references (maybe?)
        output_type (Type[DataBase]): The output type of the task, which all public models'
            .run functions must return. This must be a caikit data model type.

    Returns:
        A decorator function for the task class, registering it with caikit's core registry of
            tasks.
    """
    '''
    converted = converter.convert_to_google_style(test1)

    assert (
        converted
        == """
    
    """
    )


def test_example_usage():
    converter = CustomDocstringConverter()
    test1 = '''
    """Decorator that can be used to mark a function
    or a class as "work in progress". It will result in a warning being emitted
    when the function / class is used.

    Args:
        category: WipCategory
            Enum specifying what category of message you want to throw
        action: Action
            Enum specifying what type of action you want to take.
            Example: ERROR or WARNING

    Example Usage:

    ### Decorating class

    1. No configuration:
        @work_in_progress
        class Foo:
            pass

    2. Action and category configuration:
        @work_in_progress(action=Action.WARNING, category=WipCategory.BETA)
        class Foo:
            pass

    ### Decorating Function:

    1. No configuration:
        @work_in_progress
        def foo(*args, **kwargs):
            pass

    2. Action and category configuration:
        @work_in_progress(action=Action.WARNING, category=WipCategory.BETA)
         def foo(*args, **kwargs):
            pass

    ### Sample message:

    foo is still in the BETA phase and subject to change!

    """
    '''
    converted = converter.convert_to_google_style(test1)

    assert (
        converted
        == '''
    """Decorator that can be used to mark a function
    or a class as "work in progress". It will result in a warning being emitted
    when the function / class is used.

    Args:
        category (WipCategory): Enum specifying what category of message you
            want to throw
        action (Action): Enum specifying what type of action you want to take.
            Example: ERROR or WARNING

    Example Usage:

    ### Decorating class

    1. No configuration:
        @work_in_progress
        class Foo:
            pass

    2. Action and category configuration:
        @work_in_progress(action=Action.WARNING, category=WipCategory.BETA)
        class Foo:
            pass

    ### Decorating Function:

    1. No configuration:
        @work_in_progress
        def foo(*args, **kwargs):
            pass

    2. Action and category configuration:
        @work_in_progress(action=Action.WARNING, category=WipCategory.BETA)
         def foo(*args, **kwargs):
            pass

    ### Sample message:

    foo is still in the BETA phase and subject to change!

    """
    '''
    )


def test_raises():
    converter = CustomDocstringConverter()
    test1 = '''
    """Decorator that can be used to mark a function
    or a class as "work in progress". It will result in a warning being emitted
    when the function / class is used.

    Args:
        category: WipCategory
            Enum specifying what category of message you want to throw
        action: Action
            Enum specifying what type of action you want to take.
            Example: ERROR or WARNING

    Returns: 
        None
            This only sets the config object that is returned by `caikit.get_config()`

    Raises:
        AttributeError if a discrepancy is found between the RPC service
        descriptor and the Caikit Library CDM, which will prevent an instance of
        this class from being instantiated
    """
    '''
    converted = converter.convert_to_google_style(test1)

    assert (
        converted
        == '''
    """Decorator that can be used to mark a function
    or a class as "work in progress". It will result in a warning being emitted
    when the function / class is used.

    Args:
        category (WipCategory): Enum specifying what category of message you
            want to throw
        action (Action): Enum specifying what type of action you want to take.
            Example: ERROR or WARNING

    Returns:
         None: This only sets the config object that is returned by
            `caikit.get_config()`

    Raises:
        AttributeError if a discrepancy is found between the RPC service
        descriptor and the Caikit Library CDM, which will prevent an instance of
        this class from being instantiated
    """
    '''
    )


def test_multiword_args():
    converter = CustomDocstringConverter()
    test1 = '''
    """Decorator that can be used to mark a function
    or a class as "work in progress". It will result in a warning being emitted
    when the function / class is used.

    Args:
        category: WipCategory | list | click
            Enum specifying what category of message you want to throw
        action: Action | int
            Enum specifying what type of action you want to take.
            Example: ERROR or WARNING
        foo: a very short description
        signature: Dict[str, Type]
            module signature of parameters and types
        foo: int | int | int | int
            here's some description!
        var: foo or int or list
            more description
        kwarg: hello -> arrow
            a description

    Returns: 
        None
            This only sets the config object that is returned by `caikit.get_config()`

    """
    '''
    converted = converter.convert_to_google_style(test1)

    assert (
        converted
        == '''
    """Decorator that can be used to mark a function
    or a class as "work in progress". It will result in a warning being emitted
    when the function / class is used.

    Args:
        category (WipCategory | list | click): Enum specifying what category of
            message you want to throw
        action (Action | int): Enum specifying what type of action you want to
            take. Example: ERROR or WARNING
        foo: a very short description
        Nonesignature (Dict[str, Type]): module signature of parameters and types
        foo (int | int | int | int): here's some description!
        var (foo or int or list): more description
        kwarg (hello -> arrow): a description

    Returns:
         None: This only sets the config object that is returned by
            `caikit.get_config()`
    

    """
    '''
    )


def test_multiword_args():
    converter = CustomDocstringConverter()
    test1 = '''
    """Validate a single data item from a data stream

        Args:
            data_item: 
                A data object yielded by the stream
            data_item_number: 
                The index of the object in the stream
        """
    '''
    converted = converter.convert_to_google_style(test1)
    assert (
        converted
        == '''
    """Validate a single data item from a data stream

        Args:
            data_item: A data object yielded by the stream
            data_item_number: The index of the object in the stream
        """
    '''
    )


# Run the tests
if __name__ == "__main__":
    pytest.main([__file__])



import caikit.core


def test_caikit_core_does_not_contain_itself():
    assert not hasattr(caikit.core, "caikit.core")


def test_caikit_core_has_object_serializer():
    assert hasattr(caikit.core, "ObjectSerializer")


def test_caikit_core_has_quality_evaluation():
    assert hasattr(caikit.core, "quality_evaluation")


def test_caikit_core_has_DataValidationError():
    assert hasattr(caikit.core, "DataValidationError")


def test_caikit_core_has_error_handler():
    assert hasattr(caikit.core, "error_handler")



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This package exposes useful metadata about a method on a caikit.core module, including method
name, return type and input parameters.

This encapsulates a lot of custom logic used to parse input/output parameters, especially when it
comes to the caikit core common data model.
"""

# Standard
from typing import Any, Dict, Optional, Type
import inspect

# First Party
import alog

# Local
from . import parsers
import caikit

log = alog.use_channel("SIGNATURE")


class CaikitMethodSignature:
    """Metadata about a method on a caikit core module

    Determines the argument types and return type for a function (run, train, etc.)
    of the given caikit core module in any way possible!

    This is the most "heuristic" part of this process. Currently, there is no
    well-defined mechanism for defining the types of a module's run signature in
    the caikit.core API, so it is enforced by convention only. As such, this
    package needs to reverse engineer those conventions! It does so by walking
    through a list of candidate ways to determine the signature in order least-
    hacky to most-hacky:

    1. Look for a known type mapping based on the name of the argument
    2. Look for python type annotations
    3. Look for a default value and check its type
    4. Parse the docstring
    """

    def __init__(
        self, caikit_core_module: Type["caikit.core.ModuleBase"], method_name: str
    ):
        self._module = caikit_core_module
        self._method_name = method_name

        try:
            self._method_pointer = getattr(self._module, self._method_name)
            self._default_map = parsers.get_args_with_defaults(self._method_pointer)
            method_signature = inspect.signature(self._method_pointer)
            self._return_type = parsers.get_output_type_name(
                self._module, method_signature, self._method_pointer
            )

            self._parameters = parsers.get_argument_types(self._method_pointer)
        except AttributeError:
            log.warning(
                "Could not find method [%s] in this module",
                self.method_name,
            )
            self._return_type = None
            self._parameters = None
            self._default_map = {}

    @property
    def module(self) -> Type["caikit.core.ModuleBase"]:
        """The concrete caikit.core.ModuleBase type"""
        return self._module

    @property
    def method_name(self) -> str:
        """The name of the method on this module, e.g. 'run' or 'train'"""
        return self._method_name

    @property
    def return_type(self) -> Optional[Type]:
        """The return type annotation of the method, or None if the method does not exist"""
        return self._return_type

    @property
    def parameters(self) -> Optional[Dict[str, Type]]:
        """A dictionary of the parameter names to their types, or None if the method does not
        exist"""
        return self._parameters

    @property
    def default_parameters(self) -> Dict[str, Any]:
        """A set of all parameter names which have default values"""
        return self._default_map


class CustomSignature(CaikitMethodSignature):
    """(TBD on new class)? Need something to hold an intentionally mutated representation of a
    method signature. This represents the extra indirection that lives in the runtime, between the
    service API and the actual method. For example: .train functions return a fully constructed
    module, but the runtime will invoke .train asynchronously and instead return some handle that
    can be used to check training status."""

    def __init__(
        self,
        original_signature: CaikitMethodSignature,
        parameters: Dict[str, Type],
        return_type: Optional[Type],
    ):
        super().__init__(original_signature.module, original_signature.method_name)
        self._module = original_signature.module
        self._method_name = original_signature.method_name
        self._method_pointer = original_signature._method_pointer

        self._return_type = return_type
        self._parameters = parameters



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This library defines the taxonomy of Data Model objects and Tasks for the
entire CAIKit project. Data objects and tasks are grouped domain, making for a
three-level hierarchy for models:

problem domain -> task -> implementation

This library intentionally does NOT define any implementations, as those are
left to the domain implementation libraries.
"""

# NOTE: We do not proactively import sub-modules in the interfaces here as they
#   may contain optional dependencies



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# Standard
from tempfile import TemporaryDirectory
import os
import unittest
import uuid

# Third Party
import grpc

# Local
from caikit import get_config
from caikit.runtime.model_management.model_sizer import ModelSizer
from caikit.runtime.types.caikit_runtime_exception import CaikitRuntimeException
from tests.conftest import random_test_id, temp_config
from tests.fixtures import Fixtures


def _random_test_model_type() -> str:
    return "test-any-type-" + str(uuid.uuid4())


class TestModelSizer(unittest.TestCase):
    """This test suite tests the model sizer class"""

    def setUp(self):
        """This method runs before each test begins to run"""
        self.model_sizer = ModelSizer.get_instance()

    @staticmethod
    def _add_file(path, charsize) -> int:
        with open(path, "w") as f:
            content = ""
            for i in range(charsize):
                content += str(i)
            f.write(content)
        return os.path.getsize(path)

    def test_it_can_size_a_model_folder(self):
        """Get local model directory size"""
        with TemporaryDirectory() as d:
            total_size = TestModelSizer._add_file(os.path.join(d, "some_file"), 256)

            # add a nested file as well
            subdir = os.path.join(d, "some_dir")
            os.makedirs(subdir)
            total_size += TestModelSizer._add_file(
                os.path.join(subdir, "some_file"), 512
            )
            model_type = _random_test_model_type()
            mult = 7
            with temp_config(
                {
                    "inference_plugin": {
                        "model_mesh": {"model_size_multipliers": {model_type: mult}}
                    }
                }
            ):
                expected_size = total_size * mult
                size = self.model_sizer.get_model_size(
                    model_id=random_test_id(),
                    local_model_path=d,
                    model_type=model_type,
                )
                self.assertEqual(size, expected_size)

    def test_it_can_size_a_model_archive(self):
        """Get local model archive file size"""
        model_type = _random_test_model_type()
        mult = 42
        with temp_config(
            {
                "inference_plugin": {
                    "model_mesh": {"model_size_multipliers": {model_type: mult}}
                }
            }
        ):
            expected_size = (
                os.path.getsize(Fixtures.get_good_model_archive_path()) * mult
            )

            size = self.model_sizer.get_model_size(
                model_id=random_test_id(),
                local_model_path=Fixtures.get_good_model_archive_path(),
                model_type=model_type,
            )
            self.assertEqual(size, expected_size)

    def test_it_uses_the_default_multiplier_for_unknown_model_types(self):
        model_type = "definitely not a real type"
        expected_size = (
            os.path.getsize(Fixtures.get_good_model_archive_path())
            * get_config().inference_plugin.model_mesh.default_model_size_multiplier
        )

        size = self.model_sizer.get_model_size(
            model_id=random_test_id(),
            local_model_path=Fixtures.get_good_model_archive_path(),
            model_type=model_type,
        )
        self.assertEqual(size, expected_size)

    def test_it_throws_not_found_if_file_does_not_exist(self):
        with self.assertRaises(CaikitRuntimeException) as context:
            self.model_sizer.get_model_size(
                model_id=random_test_id(),
                local_model_path="not/a/path/to/anything",
                model_type=_random_test_model_type(),
            )
        self.assertEqual(grpc.StatusCode.NOT_FOUND, context.exception.status_code)

    def test_it_caches_model_archive_sizes(self):
        """Check that we avoid unnecessary disk access"""
        with TemporaryDirectory() as d:
            TestModelSizer._add_file(os.path.join(d, "some_file"), 256)

            model_type = _random_test_model_type()
            model_id = random_test_id()
            original_size = self.model_sizer.get_model_size(
                model_id=model_id,
                local_model_path=d,
                model_type=model_type,
            )

            TestModelSizer._add_file(os.path.join(d, "some_new_file"), 256)
            # New file is not accounted for in size, because we cached
            new_size = self.model_sizer.get_model_size(
                model_id=model_id,
                local_model_path=d,
                model_type=model_type,
            )
            self.assertEqual(original_size, new_size)


if __name__ == "__main__":
    unittest.main()



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""This file exists for backwards compatibility.
It imports from the caikit.core.exceptions package where things moved"""

# Standard
import warnings as _warnings

# Local
from caikit.core.exceptions.error_handler import *  # pylint: disable=wildcard-import,unused-wildcard-import

# 🌶️🌶️🌶️ Warn if this package is ever imported
# Allow DeprecationWarnings through if anything tries to import from `toolkit.errors`
_warnings.filterwarnings("default", category=DeprecationWarning)
# And actually warn them
_warnings.warn(
    "The caikit.toolkit.error_handler package has moved to caikit.core.exceptions",
    DeprecationWarning,
)



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# Standard
from unittest import mock
import concurrent.futures
import datetime
import re
import threading
import time

# Third Party
import grpc
import pytest

# Local
from caikit.core import MODEL_MANAGER
from caikit.core.data_model import DataStream, TrainingStatus
from caikit.core.exceptions.caikit_core_exception import (
    CaikitCoreException,
    CaikitCoreStatusCode,
)
from caikit.core.model_management.model_trainer_base import TrainingInfo
from caikit.interfaces.runtime.data_model import (
    TrainingInfoRequest,
    TrainingStatusResponse,
)
from caikit.runtime.servicers.training_management_servicer import (
    TrainingManagementServicerImpl,
)
from caikit.runtime.types.caikit_runtime_exception import CaikitRuntimeException
from sample_lib import SampleModule


@pytest.fixture
def training_pool():
    with concurrent.futures.ThreadPoolExecutor(max_workers=5) as executor:
        yield executor


@pytest.fixture
def training_management_servicer():
    return TrainingManagementServicerImpl()


class MockModelFuture:
    """Mocks up a model future that will return a 404 error after being cancelled"""

    def __init__(self) -> None:
        self._canceled = False

    def get_info(self) -> TrainingInfo:
        if not self._canceled:
            return TrainingInfo(status=TrainingStatus.RUNNING)
        else:
            raise CaikitCoreException(
                status_code=CaikitCoreStatusCode.NOT_FOUND, message="Training not found"
            )

    def cancel(self):
        self._canceled = True


def test_training_runs(training_management_servicer, training_pool):

    # Create a future and set it in the training manager
    event = threading.Event()
    model_future = MODEL_MANAGER.train(
        SampleModule,
        DataStream.from_iterable([]),
        wait_event=event,
    )

    # send a request, check it's running
    request = TrainingInfoRequest(training_id=model_future.id).to_proto()
    response = training_management_servicer.GetTrainingStatus(request, context=None)
    assert response.state == TrainingStatus.RUNNING.value

    event.set()
    model_future.wait()

    # Ensure it's now done
    request = TrainingInfoRequest(training_id=model_future.id).to_proto()
    response = training_management_servicer.GetTrainingStatus(request, context=None)
    assert response.state == TrainingStatus.COMPLETED.value


def test_training_timestamps(training_management_servicer, training_pool):
    """Check that the submission and completion timestamps are set correctly from the model
    future."""
    # Create a future and set it in the training manager
    event = threading.Event()
    model_future = MODEL_MANAGER.train(
        SampleModule,
        DataStream.from_iterable([]),
        wait_event=event,
    )

    # send a request, check that the submission timestamp was set
    request = TrainingInfoRequest(training_id=model_future.id).to_proto()
    response = training_management_servicer.GetTrainingStatus(request, context=None)
    response_dm: TrainingStatusResponse = TrainingStatusResponse.from_proto(response)
    assert response_dm.submission_timestamp is not None
    assert (
        datetime.datetime.now() - response_dm.submission_timestamp
        < datetime.timedelta(milliseconds=50)
    )
    assert response_dm.completion_timestamp is None

    event.set()
    # Wait for training to complete, but _don't_ interact with the future.
    # This simulates async training on the runtime
    for i in range(1000):
        response = training_management_servicer.GetTrainingStatus(request, context=None)
        if response.state == TrainingStatus.COMPLETED.value:
            break

    # Ensure completion timestamp is set
    response_dm: TrainingStatusResponse = TrainingStatusResponse.from_proto(response)
    assert response_dm.completion_timestamp is not None
    assert (
        datetime.datetime.now() - response_dm.completion_timestamp
        < datetime.timedelta(milliseconds=50)
    )
    assert response_dm.completion_timestamp > response_dm.submission_timestamp


def test_training_cannot_cancel_on_completed_training(training_management_servicer):
    # Create a future and set it in the training manager
    event = threading.Event()
    model_future = MODEL_MANAGER.train(
        SampleModule,
        DataStream.from_iterable([]),
        wait_event=event,
    )

    # send a request, check it's not errored
    request = TrainingInfoRequest(training_id=model_future.id).to_proto()
    response = training_management_servicer.GetTrainingStatus(request, context=None)
    assert response.state != TrainingStatus.ERRORED.value

    event.set()
    model_future.wait()

    # cancel the training request, but this won't change its status as it was finished
    training_management_servicer.CancelTraining(request, context=None)
    response = training_management_servicer.GetTrainingStatus(request, context=None)
    assert response.state == TrainingStatus.COMPLETED.value


def test_training_cancel_on_correct_id(training_management_servicer):
    # Create a training future for first model with a long runtime
    # NOTE: We cannot use a single log time.sleep or a threading.Event().wait
    #   here because those are not actually destroyable via DestroyableThread.
    #   Instead, we use the sleep_time argument that runs a loop of short sleeps
    #   and then we verify below that the training duration was far less than
    #   this configured sleep time.
    start_event = threading.Event()
    model_future_1 = MODEL_MANAGER.train(
        SampleModule,
        DataStream.from_iterable([]),
        sleep_time=100,
        start_event=start_event,
    )

    # Wait until the training has started to ensure it is interrupted in flight
    start_event.wait()

    request_1 = TrainingInfoRequest(training_id=model_future_1.id).to_proto()
    response_1 = training_management_servicer.GetTrainingStatus(request_1, context=None)
    assert response_1.state == TrainingStatus.RUNNING.value

    # Model 2 has no wait event, should proceed to complete training
    model_future_2 = MODEL_MANAGER.train(
        SampleModule,
        DataStream.from_iterable([1, 2, 3]),
    )

    # Cancel first training
    request_1 = TrainingInfoRequest(training_id=model_future_1.id).to_proto()
    training_management_servicer.CancelTraining(request_1, context=None)

    response_1 = training_management_servicer.GetTrainingStatus(request_1, context=None)
    assert response_1.state == TrainingStatus.CANCELED.value

    # Make sure the model future completes without the blocking event being set
    start_time = time.time()
    model_future_1.wait()
    # Sanity check that the model did not wait for anywhere close to the full
    # 100 seconds
    wait_time = time.time() - start_time
    assert wait_time < 5
    assert (
        training_management_servicer.GetTrainingStatus(request_1, context=None).state
        == TrainingStatus.CANCELED.value
    )

    # training number 2 should still complete
    request_2 = TrainingInfoRequest(training_id=model_future_2.id).to_proto()
    response_2 = training_management_servicer.GetTrainingStatus(request_2, context=None)
    assert response_2.state == TrainingStatus.COMPLETED.value


def test_training_cancel_on_mock_model_future(training_management_servicer):
    # Patch in our mock model future
    with mock.patch.object(MODEL_MANAGER, "get_model_future") as mock_gmf:
        mock_gmf.return_value = MockModelFuture()

        # Check that we get "running" status
        info_request = TrainingInfoRequest(training_id="anything").to_proto()
        info_response = training_management_servicer.GetTrainingStatus(
            info_request, context=None
        )
        assert info_response.state == TrainingStatus.RUNNING.value

        # Make sure a cancel returns "canceled"
        cancel_response = training_management_servicer.CancelTraining(
            info_request, context=None
        )
        assert cancel_response.state == TrainingStatus.CANCELED.value


def test_training_complete_status(training_management_servicer, training_pool):

    # Create a future and set it in the training manager
    model_future = MODEL_MANAGER.train(
        SampleModule,
        DataStream.from_iterable([]),
    )

    # Block until train is done
    model_future.wait()

    # send a request
    request = TrainingInfoRequest(training_id=model_future.id).to_proto()
    response = training_management_servicer.GetTrainingStatus(request, context=None)

    assert re.match("<class '.*TrainingStatusResponse'>", str(type(response)))
    assert response.state == TrainingStatus.COMPLETED.value


def test_training_status_incorrect_id(training_management_servicer):
    # send a request with a training id that doesn't exist
    request = TrainingInfoRequest(training_id="some_random_id").to_proto()
    with pytest.raises(CaikitRuntimeException) as context:
        training_management_servicer.GetTrainingStatus(request, context=None)

    assert context.value.status_code == grpc.StatusCode.NOT_FOUND
    assert (
        f"Unknown training_id: some_random_id" in context.value.message
    )  # message set by local_model_trainer.get_model_future


def test_training_raises_when_cancel_on_incorrect_id(training_management_servicer):
    # Create a future and set it in the training manager
    event = threading.Event()
    model_future = MODEL_MANAGER.train(
        SampleModule,
        DataStream.from_iterable([]),
        wait_event=event,
    )

    # send a request, check it's not errored
    request = TrainingInfoRequest(training_id=model_future.id).to_proto()
    response = training_management_servicer.GetTrainingStatus(request, context=None)
    assert response.state != TrainingStatus.ERRORED.value

    event.set()
    model_future.wait()

    # cancel a training with wrong training id
    cancel_request = TrainingInfoRequest(training_id="some_random_id").to_proto()
    with pytest.raises(CaikitRuntimeException) as context:
        training_management_servicer.CancelTraining(cancel_request, context=None)

    assert context.value.status_code == grpc.StatusCode.NOT_FOUND
    assert (
        f"Unknown training_id: some_random_id" in context.value.message
    )  # message set by local_model_trainer.get_model_future


def test_training_fails(training_management_servicer, training_pool):
    model_future = MODEL_MANAGER.train(
        SampleModule,
        DataStream.from_iterable([]),
        batch_size=SampleModule.POISON_PILL_BATCH_SIZE,
    )
    model_future.wait()

    # send a request
    request = TrainingInfoRequest(training_id=model_future.id).to_proto()
    response = training_management_servicer.GetTrainingStatus(request, context=None)

    assert response.state == TrainingStatus.ERRORED.value



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for sync -> async wrappers"""
# Standard
from concurrent.futures import ThreadPoolExecutor
import asyncio
import threading
import time

# Third Party
import pytest

# Local
from caikit.core.toolkit.sync_to_async import async_wrap_iter

## Helpers #####################################################################


class IterableHelper:
    def __init__(self, iterable, delay=0, raise_element=None):
        self.iterable = iterable
        self.delay = delay
        self.raise_element = raise_element
        self.idx = 0
        self.iter = iter(self.iterable)

    def __iter__(self):
        return self

    def __next__(self):
        if self.delay:
            time.sleep(self.delay)
        next_element = next(self.iter)
        if self.idx == self.raise_element:
            raise RuntimeError("Raising")
        self.idx += 1
        return next_element


async def azip(*aiterables):
    aiterators = [aiterable.__aiter__() for aiterable in aiterables]
    while aiterators:
        results = await asyncio.gather(
            *[ait.__anext__() for ait in aiterators],
            return_exceptions=True,
        )
        if any(isinstance(result, StopAsyncIteration) for result in results):
            return
        yield tuple(results)


## Tests #######################################################################


@pytest.mark.asyncio
async def test_async_wrap_iter_happy():
    """Make sure a simple iterable can be wrapped as async and round trip
    cleanly
    """
    lst = [1, 2, 3, 4]
    async_round_trip = [x async for x in async_wrap_iter(lst)]
    assert lst == async_round_trip


@pytest.mark.asyncio
async def test_async_wrap_iter_concurrent():
    """Make sure that multiple wrapped synchronous iterators can interleve in
    the event loop
    """
    lst1 = [1, 2, 3]
    lst2 = [4, 5, 6]
    iter1 = IterableHelper(lst1)
    iter2 = IterableHelper(lst2)
    aiter1 = async_wrap_iter(iter1)
    aiter2 = async_wrap_iter(iter2)
    zipped = [elt async for elt in azip(aiter1, aiter2)]
    assert zipped == list(zip(lst1, lst2))


@pytest.mark.asyncio
async def test_async_wrap_iter_exception_propagation():
    """Make sure exceptions thrown during iteration are raised in the parent
    async context
    """
    with pytest.raises(RuntimeError):
        async for _ in async_wrap_iter(IterableHelper([1, 2, 3], raise_element=1)):
            pass


@pytest.mark.asyncio
async def test_async_wrap_iter_shared_executor():
    """Make sure that a shared ThreadPoolExecutor can be used to manage the work
    without exhausting thread counts
    """
    lst = [1, 2, 3, 4]
    pool = ThreadPoolExecutor(max_workers=1)

    start_event = threading.Event()
    finish_event = threading.Event()

    def slow_task():
        start_event.set()
        finish_event.wait()

    # Start the slow task and make sure it's started
    slow_future = pool.submit(slow_task)
    start_event.wait()

    # Use the same pool to start the async iteration
    agen = async_wrap_iter(lst, pool)

    # Make sure the async iteration is blocked
    assert pool._work_queue.qsize() == 1

    # Wait for the slow task to finish
    finish_event.set()
    slow_future.result()

    # Make sure the async iteration is unblocked
    assert pool._work_queue.qsize() == 0
    async_round_trip = [x async for x in agen]
    assert lst == async_round_trip



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# Standard
import threading
import time
import unittest

# Local
from caikit.runtime.types.aborted_exception import AbortedException
from caikit.runtime.work_management.abortable_action import AbortableAction
from caikit.runtime.work_management.rpc_aborter import RpcAborter
from tests.fixtures import Fixtures


class TestAbortableAction(unittest.TestCase):
    """This test suite tests the abortable action class"""

    def setUp(self):
        """This method runs before each test begins to run"""
        self.rpc_context = Fixtures.build_context()
        self.aborter = RpcAborter(self.rpc_context)

    def test_it_can_run_a_function(self):
        expected_result = "test-any-result"
        action = AbortableAction(self.aborter, lambda *args, **kwargs: expected_result)
        result = action.do()
        self.assertEqual(expected_result, result)

    def test_it_raises_if_the_rpc_has_already_terminated(self):
        action = AbortableAction(self.aborter, lambda *args, **kwargs: None)
        self.rpc_context.cancel()

        with self.assertRaises(AbortedException) as context:
            action.do()

    def test_it_raises_if_the_function_raises(self):
        expected_exception = ValueError("test-any-error")

        def thrower():
            raise expected_exception

        action = AbortableAction(self.aborter, thrower)
        with self.assertRaises(ValueError) as ctx:
            action.do()

        self.assertEqual(expected_exception, ctx.exception)

    def test_it_raises_if_the_rpc_is_terminated_mid_function(self):
        infinite_function_has_started = threading.Event()

        def infinite_function():
            infinite_function_has_started.set()
            while True:
                time.sleep(0.1)

        action = AbortableAction(self.aborter, infinite_function)

        def inner_test_thread():
            with self.assertRaises(AbortedException) as context:
                action.do()

        thread = threading.Thread(target=inner_test_thread)
        thread.start()
        infinite_function_has_started.wait()

        self.rpc_context.cancel()
        thread.join(5)
        self.assertFalse(thread.is_alive())



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Most logic interacting with models.  Can load, etc.
"""

# Standard
from contextlib import contextmanager
from io import BytesIO
from threading import Lock
from typing import Dict, Optional, Type, Union
import os
import tempfile
import zipfile

# First Party
import alog

# Local
from ..interfaces.common.data_model.stream_sources import S3Path
from .exceptions import error_handler
from .model_management import (
    ModelFinderBase,
    ModelInitializerBase,
    ModelTrainerBase,
    model_finder_factory,
    model_initializer_factory,
    model_trainer_factory,
)
from .modules.base import ModuleBase
from .registries import module_registry
from .toolkit.factory import Factory, FactoryConstructible
from caikit.config import get_config

log = alog.use_channel("MDLMNG")
error = error_handler.get(log)

# restrict functions that are imported so we don't pollute the base module namespace
__all__ = [
    "get_valid_module_ids",
    "ModelManager",
]


def get_valid_module_ids():
    """Get a dictionary mapping all module IDs to the string names of the
    implementing classes.
    """
    return {
        module_id: model_class.__name__
        for module_id, model_class in module_registry().items()
    }


class ModelManager:
    """Manage the models or resources for library."""

    def __init__(self):
        """Initialize ModelManager."""
        # Map to store module caches, to be used for singleton model lookups
        self._singleton_module_cache = {}
        self._trainers = {}
        self._finders = {}
        self._initializers = {}
        self.__singleton_lock = Lock()

    def initialize_components(self):
        """Proactively initialize all configured trainer/finder/initializer
        component instances. This is a separate call to enable explicit config.
        """
        # Initialize all configured components
        mm_config = get_config().model_management
        for trainer in mm_config.get("trainers", {}):
            self.get_trainer(trainer)
        for finder in mm_config.get("finders", {}):
            self.get_finder(finder)
        for initializer in mm_config.get("initializers", {}):
            self.get_initializer(initializer)

    ## Public ##################################################################

    def train(
        self,
        module: Union[Type[ModuleBase], str],
        *args,
        trainer: Union[str, ModelTrainerBase] = "default",
        save_path: Optional[Union[str, S3Path]] = None,
        save_with_id: bool = False,
        model_name: Optional[str] = None,
        wait: bool = False,
        **kwargs,
    ) -> ModelTrainerBase.ModelFutureBase:
        """Train an instance of the given module with the given args and kwargs
        using the given trainer.

        Each module's train function encapsulates the code needed to perform the
        training locally. This top-level train function provides the wrapper
        functionality to delegate the execution of the module's train function
        to an alternate framework using a ModelTrainerBase. It also allows
        training to be launched asynchronously.

        Args:
            module (Union[Type[ModuleBase], str]): The module class or guid for
                the module to train
            *args: Additional positional args to pass through to the module's
                train function

        Kwargs:
            trainer (Union[str, ModelTrainerBase]): The trainer to use. If given
                as a string, this is a key in the global config at
                model_management.trainers.
            save_path (Optional[Union[str, S3Path]]): Base path where the model should be
                saved (may be relative to a remote trainer's filesystem, or link to S3
                storage)
            save_with_id (bool): Inject the training ID into the save path for
                the output model
            model_name (Optional[str]): Name of model that will be appended
                to the end of the save_path
            wait (bool): Wait for training to complete before returning
            **kwargs: Additional keyword arguments to pass through to the
                modules's train function

        Returns:
            model_future (ModelTrainerBase.ModelFutureBase): The future handle
                to the model which holds the status of the in-flight training.
        """
        # Resolve the module class
        if isinstance(module, str):
            module_id = module
            module = module_registry().get(module_id)
            error.value_check(
                "<COR00469102E>",
                module is not None,
                "Unable to train unknown module {}",
                module_id,
            )
        error.subclass_check("<COR05418775E>", module, ModuleBase)

        # Get the trainer to use
        trainer: ModelTrainerBase = self.get_trainer(trainer)

        # Start the training
        with alog.ContextTimer(log.debug, "Started training in: "):
            model_future = trainer.train(
                module,
                *args,
                save_path=save_path,
                save_with_id=save_with_id,
                model_name=model_name,
                **kwargs,
            )
            log.debug(
                "Started training %s with save path %s",
                model_future.id,
                model_future.save_path,
            )

        # If requested, wait for the future to complete
        if wait:
            log.debug("Waiting for training %s to complete", model_future.id)
            with alog.ContextTimer(
                log.debug, "Finished training %s in: ", model_future.id
            ):
                model_future.wait()

        # Return a handle to the training
        return model_future

    def get_model_future(
        self,
        training_id: str,
    ) -> ModelTrainerBase.ModelFutureBase:
        """Get the future handle to an in-progress training

        Args:
            training_id (str): The ID string from the original training
                submission's ModelFuture

        Returns:
            model_future (ModelTrainerBase.ModelFutureBase): The future handle
                to the model which holds the status of the in-flight training.
        """
        try:
            trainer = self.get_trainer(ModelTrainerBase.get_trainer_name(training_id))

        # Fall back to the default trainer to try to find this ID
        except ValueError:
            trainer = self.get_trainer("default")

        return trainer.get_model_future(training_id)

    def load(
        self,
        module_path: Union[str, BytesIO, bytes],
        *,
        load_singleton: bool = False,
        finder: Union[str, ModelFinderBase] = "default",
        initializer: Union[str, ModelInitializerBase] = "default",
        **kwargs,
    ):
        """Load a model and return an instantiated object on which we can run
        inference.

        Args:
            module_path (str | BytesIO | bytes): A module path (identifier) to
                one of the following:
                1. A directory containing a yaml config file in the top level.
                2. A zip archive containing either a yaml config file in the
                    top level when extracted, or a directory containing a yaml
                    config file in the top level.
                3. A BytesIO object corresponding to a zip archive containing
                    either a yaml config file in the top level when extracted,
                    or a directory containing a yaml config file in the top
                    level.
                4. A bytes object corresponding to a zip archive containing
                    either a yaml config file in the top level when extracted,
                    or a directory containing a yaml config file in the top
                    level.
                5. A string that is understood by the configured
                    finder/initializer

        Kwargs:
            load_singleton (bool): Load this model as a singleton
            finder (Union[str, ModelFinderBase]): Finder to use when loading
                this model. If passed as a string, this names the finder in the
                global config model_management.finders section.
            initializer (Union[str, ModelInitializerBase]): Loader to use when
                initializint this model. If passed as a string, this is the name
                of the initializer in the global
                config model_management.initializers section.

        Returns:
            model (ModuleBase) Model object that is loaded, configured, and
                ready for prediction.
        """
        error.type_check("<COR98255724E>", bool, load_singleton=load_singleton)

        # This allows a user to load their own model (e.g. model saved to disk)
        load_path = get_config().load_path
        if load_path is not None and isinstance(module_path, str):
            if not os.path.exists(module_path):
                full_module_path = os.path.join(load_path, module_path)
                if os.path.exists(full_module_path):
                    module_path = full_module_path

        # Ensure that we have a loadable directory.
        error.type_check("<COR98255419E>", str, BytesIO, bytes, module_path=module_path)
        if isinstance(module_path, str):
            # Ensure this path is operating system correct if it isn't already.
            module_path = os.path.normpath(module_path)
        # If we have bytes, convert to a buffer, since we already handle in memory binary streams.
        elif isinstance(module_path, bytes):
            module_path = BytesIO(module_path)

        # Now that we have a file like object | str we can try to load as an archive.
        if zipfile.is_zipfile(module_path):
            return self._load_from_zipfile(
                module_path, load_singleton, finder, initializer, **kwargs
            )
        try:
            return self._do_load(
                module_path, load_singleton, finder, initializer, **kwargs
            )
        except FileNotFoundError:
            error(
                "<COR80419785E>",
                FileNotFoundError(
                    "Module load path `{}` does not contain a `config.yml` file.".format(
                        module_path
                    )
                ),
            )

    def extract(
        self, zip_path: str, model_path: str, force_overwrite: bool = False
    ) -> str:
        """Method to extract a downloaded archive to a specified directory.

        Args:
            zip_path (str): Location of .zip file to extract.
            model_path (str): Model directory where the archive should be
                unzipped unzipped.
            force_overwrite: bool (Defaults to false)
                Force an overwrite to model_path, even if the folder exists
        Returns:
            str: Output path where the model archive is unzipped.
        """
        model_path = os.path.abspath(model_path)

        # skip if force_overwrite disabled and path already exists
        if not force_overwrite and os.path.exists(model_path):
            log.info(
                "INFO: Skipped extraction. Archive already extracted in directory: %s",
                model_path,
            )
            return model_path

        with zipfile.ZipFile(zip_path, "r") as zip_f:
            zip_f.extractall(model_path)

        # path to model
        return model_path

    def resolve_and_load(
        self, path_or_name_or_model_reference: Union[str, ModuleBase], **kwargs
    ):
        """Try our best to load a model, given a path or a name. Simply returns any loaded model
        passed in. This exists to ease the burden on workflow developers who need to accept
        individual modules in their API, where users may have references to custom models or may
        only have the ability to give the name of a stock model.

        Args:
            path_or_name_or_model_reference (str, ModuleBase): Either a
                - Path to a model on disk
                - Name of a model that the catalog knows about
                - Loaded module
            **kwargs: Any keyword arguments to pass along to ModelManager.load()
                        or ModelManager.download()
                e.g. parent_dir

        Returns:
            A loaded module
        Examples:
            >>> stock_syntax_model = manager.resolve_and_load('syntax_izumo_en_stock')
            >>> local_categories_model = manager.resolve_and_load('path/to/categories/model')
            >>> some_custom_model = manager.resolve_and_load(some_custom_model)
        """
        error.type_check(
            "<COR50266694E>",
            str,
            ModuleBase,
            path_or_name_or_model_reference=path_or_name_or_model_reference,
        )

        # If this is already a module, we're good to go
        if isinstance(path_or_name_or_model_reference, ModuleBase):
            log.debug("Returning model %s directly", path_or_name_or_model_reference)
            return path_or_name_or_model_reference

        # Otherwise, this could either be a path on disk or some name of a model that our catalog
        # can resolve and fetch
        if os.path.isdir(path_or_name_or_model_reference):
            # Try to load from path
            log.debug(
                "Attempting to load model from path %s", path_or_name_or_model_reference
            )
            return self.load(path_or_name_or_model_reference, **kwargs)

        error(
            "<COR50207495E>",
            ValueError(
                "could not find model with name `{}`".format(
                    path_or_name_or_model_reference
                )
            ),
        )

    def get_singleton_model_cache_info(self):
        """Returns information about the singleton cache in {hash: module type} format

        Returns:
            Dict[str, type]: A dictionary of model hashes to model types
        """
        return {k: type(v) for k, v in self._singleton_module_cache.items()}

    def clear_singleton_cache(self):
        """Clears the cache of singleton models. Useful to release references of models, as long as
        you know that they are no longer held elsewhere and you won't be loading them again.

        Returns:
            None
        """
        with self.__singleton_lock:
            self._singleton_module_cache.clear()

    def get_trainer(self, trainer: Union[str, ModelTrainerBase]) -> ModelTrainerBase:
        """Get the configured model trainer or the one passed by value"""
        return self._get_component(
            component=trainer,
            component_dict=self._trainers,
            component_factory=model_trainer_factory,
            component_name="trainer",
            component_cfg=get_config().model_management.trainers,
            component_type=ModelTrainerBase,
        )

    def get_finder(self, finder: Union[str, ModelFinderBase]) -> ModelFinderBase:
        """Get the configured model finder or the one passed by value"""
        return self._get_component(
            component=finder,
            component_dict=self._finders,
            component_factory=model_finder_factory,
            component_name="finder",
            component_cfg=get_config().model_management.finders,
            component_type=ModelFinderBase,
        )

    def get_initializer(
        self, initializer: Union[str, ModelInitializerBase]
    ) -> ModelInitializerBase:
        """Get the configured model initializer or the one passed by value"""
        return self._get_component(
            component=initializer,
            component_dict=self._initializers,
            component_factory=model_initializer_factory,
            component_name="initializer",
            component_cfg=get_config().model_management.initializers,
            component_type=ModelInitializerBase,
        )

    ## Implementation Details ##################################################

    def _do_load(self, module_path, load_singleton, finder, initializer, **kwargs):
        """Load a model from a directory.

        Args:
            module_path (str): Path to directory. At the top level of directory
                is `config.yml` which holds info about the model.
            load_singleton (bool): Load this model as a singleton
            finder (Union[str, ModelFinderBase]): Finder to use when loading
                this model. If passed as a string, this names the finder in the
                global config model_management.finders section.
            initializer (Union[str, ModelInitializerBase]): Loader to use when
                loading this model. If passed as a string, this is the name of
                the initializer in the global
                config model_management.initializers section.

        Returns:
            subclass of caikit.core.modules.ModuleBase: Model object that is
                loaded, configured, and ready for prediction.
        """
        with self._singleton_lock(load_singleton):
            if singleton_entry := (
                load_singleton and self._singleton_module_cache.get(module_path)
            ):
                log.debug("Found %s in the singleton cache", module_path)
                return singleton_entry

            # Use the given finder to try to find the module config for this
            # module_path
            #
            # NOTE: This will lazily construct named finders if needed
            log.debug("Attempting to find [%s] with finder %s", module_path, finder)
            finder = self.get_finder(finder)
            log.debug2("Finder type: %s", finder.name)
            model_config = finder.find_model(module_path, **kwargs)
            error.value_check(
                "<COR92173495E>",
                model_config is not None,
                "Unable to find a ModuleConfig for {}",
                module_path,
            )

            # Use the given initializer to try to load the model
            #
            # NOTE: This will lazily construct named initializers if needed
            log.debug(
                "Attempting to initialize [%s] with initializer %s",
                module_path,
                initializer,
            )
            initializer = self.get_initializer(initializer)
            log.debug2("Initializer type: %s", initializer.name)
            loaded_model = initializer.init(model_config, **kwargs)
            error.value_check(
                "<COR50207494E>",
                loaded_model is not None,
                "Unable to load model from {} with MODULE_ID {}",
                module_path,
                model_config.module_id,
            )

            # If loading as a singleton, populate the cache
            if load_singleton:
                self._singleton_module_cache[module_path] = loaded_model

            # Return successfully!
            return loaded_model

    def _load_from_zipfile(
        self, module_path, load_singleton, finder, initializer, **kwargs
    ):
        """Load a model from a zip archive.

        Args:
            module_path (str): Path to directory. At the top level of directory
                is `config.yml` which holds info about the model.
            load_singleton (bool): Load this model as a singleton
            finder (Union[str, ModelFinderBase]): Finder to use when loading
                this model. If passed as a string, this names the finder in the
                global config model_management.finders section.
            initializer (Union[str, ModelInitializerBase]): Loader to use when
                loading this model. If passed as a string, this is the name of
                the initializer in the global
                config model_management.initializers section.

        Returns:
            subclass of caikit.core.modules.ModuleBase: Model object that is
                loaded, configured, and ready for prediction.
        """
        with tempfile.TemporaryDirectory() as extract_path:
            with zipfile.ZipFile(module_path, "r") as zip_f:
                zip_f.extractall(extract_path)
            # Depending on the way the zip archive is packaged, out temp directory may unpack
            # to files directly, or it may unpack to a (single) directory containing the files.
            # We expect the former, but fall back to the second if we can't find the config.
            try:
                model = self._do_load(
                    extract_path, load_singleton, finder, initializer, **kwargs
                )
            # NOTE: Error handling is a little gross here, the main reason being that we
            # only want to log to error() if something is fatal, and there are a good amount
            # of things that can go wrong in this process.
            except FileNotFoundError:

                def get_full_path(folder_name):
                    return os.path.join(extract_path, folder_name)

                # Get the contained directories. Omit anything starting with __ to avoid
                # accidentally traversing compression artifacts, e.g., __MACOSX.
                nested_dirs = [
                    get_full_path(f)
                    for f in os.listdir(extract_path)
                    if os.path.isdir(get_full_path(f)) and not f.startswith("__")
                ]
                # If we have multiple dirs, something is probably wrong - this doesn't look
                # like a simple level of nesting as a result of creating the zip.
                if len(nested_dirs) != 1:
                    error(
                        "<COR06761097E>",
                        FileNotFoundError(
                            "Unable to locate archive config due to nested dirs"
                        ),
                    )
                # Otherwise, try again. If we fail again stop, because the zip creation should only
                # create one potential extra layer of nesting around the model directory.
                try:
                    model = self._do_load(
                        nested_dirs[0], load_singleton, finder, initializer, **kwargs
                    )
                except FileNotFoundError:
                    error(
                        "<COR84410081E>",
                        FileNotFoundError(
                            "Unable to locate archive config within top two levels of {}".format(
                                module_path
                            )
                        ),
                    )
        return model

    @contextmanager
    def _singleton_lock(self, load_singleton: bool):
        """Helper contextmanager that will only lock the singleton cache if this
        load is a singleton load
        """
        if load_singleton:
            with self.__singleton_lock:
                yield
        else:
            yield

    @staticmethod
    def _get_component(
        component: Union[str, FactoryConstructible],
        component_dict: Dict[str, FactoryConstructible],
        component_factory: Factory,
        component_name: str,
        component_cfg: dict,
        component_type: type,
    ) -> FactoryConstructible:
        """Common logic for resolving components from config

        NOTE: This is done lazily to avoid relying on import order and to allow
            for dynamic config changes
        """
        error.type_check(
            "<COR45466249E>", str, component_type, **{component_name: component}
        )
        if isinstance(component, component_type):
            return component
        if component not in component_dict:
            cfg = component_cfg.get(component)
            error.value_check(
                "<COR55057389E>",
                isinstance(cfg, dict),
                "Unknown {}: {}",
                component_name,
                component,
            )
            component_dict[component] = component_factory.construct(cfg, component)
        return component_dict[component]



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# Standard
from contextlib import contextmanager
from importlib.metadata import version
from typing import Any, Dict, Iterable, Optional, Set, Union
import itertools
import traceback

# Third Party
from google.protobuf.descriptor import FieldDescriptor
from google.protobuf.message import Message as ProtobufMessage
from grpc import ServicerContext, StatusCode
from prometheus_client import Counter, Summary

# First Party
import alog

# Local
from caikit import get_config
from caikit.core import ModuleBase
from caikit.core.data_model import DataBase, DataStream
from caikit.core.signature_parsing import CaikitMethodSignature
from caikit.runtime.metrics.rpc_meter import RPCMeter
from caikit.runtime.model_management.model_manager import ModelManager
from caikit.runtime.service_factory import ServicePackage
from caikit.runtime.service_generation.rpcs import TaskPredictRPC
from caikit.runtime.types.caikit_runtime_exception import CaikitRuntimeException
from caikit.runtime.utils.import_util import clean_lib_names
from caikit.runtime.utils.servicer_util import (
    build_caikit_library_request_dict,
    build_proto_response,
    build_proto_stream,
    get_metadata,
    validate_data_model,
)
from caikit.runtime.work_management.abortable_action import AbortableAction
from caikit.runtime.work_management.rpc_aborter import RpcAborter

PREDICT_RPC_COUNTER = Counter(
    "predict_rpc_count",
    "Count of global predict-managed RPC calls",
    ["grpc_request", "code", "model_id"],
)
PREDICT_FROM_PROTO_SUMMARY = Summary(
    "predict_from_proto_duration_seconds",
    "Histogram of predict request unmarshalling duration (in seconds)",
    ["grpc_request", "model_id"],
)
PREDICT_CAIKIT_LIBRARY_SUMMARY = Summary(
    "predict_caikit_library_duration_seconds",
    "Histogram of predict Caikit Library run duration (in seconds)",
    ["grpc_request", "model_id"],
)
PREDICT_TO_PROTO_SUMMARY = Summary(
    "predict_to_proto_duration_seconds",
    "Histogram of predict response marshalling duration (in seconds)",
    ["grpc_request", "model_id"],
)

log = alog.use_channel("GP-SERVICR-I")

# Protobuf non primitives
# Ref: https://developers.google.com/protocol-buffers/docs/reference/cpp/google.protobuf.descriptor
NON_PRIMITIVE_TYPES = [FieldDescriptor.TYPE_MESSAGE, FieldDescriptor.TYPE_ENUM]


class GlobalPredictServicer:
    """This class contains RPC calls affiliated with the Caikit Runtime that are
    not a part of the Model Runtime proto definition.  They will be serviced by
    mocking the particular RPC (based on the message type), leveraging the
    CaikitRuntimeServicerMock to return the appropriate mock response for a
    given request
    """

    # Invocation metadata key for the model ID, provided by Model Mesh
    MODEL_MESH_MODEL_ID_KEY = "mm-model-id"

    # Input size in code points, provided by orchestrator
    INPUT_SIZE_KEY = "input-length"

    def __init__(
        self,
        inference_service: ServicePackage,
        use_abortable_threads: bool = get_config().runtime.use_abortable_threads,
    ):
        self._started_metering = False
        self._model_manager = ModelManager.get_instance()
        self._rpc_meter = None
        if get_config().runtime.metering.enabled:
            self._started_metering = True
            self.rpc_meter = RPCMeter()
            log.info(
                "<RUN76773775I>",
                "Metering is enabled, to disable set `metering.enabled` in config to false",
            )
        else:
            log.info(
                "<RUN76773776I>",
                "Metering is disabled, to enable set `metering.enabled` in config to true",
            )

        self.use_abortable_threads = use_abortable_threads
        self._inference_service = inference_service
        # Validate that the Caikit Library CDM is compatible with our service descriptor
        validate_data_model(self._inference_service.descriptor)
        log.info("<RUN76773778I>", "Validated Caikit Library CDM successfully")

        # Duplicate code in global_train_servicer
        # pylint: disable=duplicate-code
        library = clean_lib_names(get_config().runtime.library)[0]
        try:
            lib_version = version(library)
        except Exception:  # pylint: disable=broad-exception-caught
            lib_version = "unknown"

        log.info(
            "<RUN76884779I>",
            "Constructed inference service for library: %s, version: %s",
            library,
            lib_version,
        )
        super()

    def Predict(
        self,
        request: Union[ProtobufMessage, Iterable[ProtobufMessage]],
        context: ServicerContext,
        caikit_rpc: TaskPredictRPC,
        *_,
        **__,
    ) -> Union[ProtobufMessage, Iterable[ProtobufMessage]]:
        """Global predict RPC -- Mocks the invocation of a Caikit Library module.run()
        method for a loaded Caikit Library model

        Args:
            request (ProtobufMessage):
                A deserialized RPC request message
            context (ServicerContext):
                Context object (contains request metadata, etc)

        Returns:
            response (Union[ProtobufMessage, Iterable[ProtobufMessage]]):
                A Caikit Library data model response object
        """
        # Make sure the request has a model before doing anything
        model_id = get_metadata(context, self.MODEL_MESH_MODEL_ID_KEY)
        request_name = caikit_rpc.request.name

        with self._handle_predict_exceptions(model_id, request_name):
            with alog.ContextLog(
                log.debug, "GlobalPredictServicer.Predict:%s", request_name
            ):
                # Retrieve the model from the model manager
                log.debug("<RUN52259029D>", "Retrieving model '%s'", model_id)
                model = self._model_manager.retrieve_model(model_id)
                model_class = type(model)

                # Little hackity hack: Calling _verify_model_task upfront here as well to
                # short-circuit requests where the model is _totally_ unsupported
                self._verify_model_task(model)

                # Unmarshall the request object into the required module run argument(s)
                with PREDICT_FROM_PROTO_SUMMARY.labels(
                    grpc_request=request_name, model_id=model_id
                ).time():
                    inference_signature = model_class.get_inference_signature(
                        input_streaming=caikit_rpc.input_streaming,
                        output_streaming=caikit_rpc.output_streaming,
                    )
                    if not inference_signature:
                        raise CaikitRuntimeException(
                            StatusCode.INVALID_ARGUMENT,
                            f"Model class {model_class} does not support {caikit_rpc.name}",
                        )
                    if caikit_rpc.input_streaming:
                        caikit_library_request = (
                            self._build_caikit_library_request_stream(
                                request, inference_signature, caikit_rpc
                            )
                        )
                    else:
                        caikit_library_request = build_caikit_library_request_dict(
                            request,
                            inference_signature,
                        )
                response = self.predict_model(
                    request_name,
                    model_id,
                    inference_func_name=inference_signature.method_name,
                    aborter=RpcAborter(context) if self.use_abortable_threads else None,
                    **caikit_library_request,
                )

                # Marshall the response to the necessary return type
                with PREDICT_TO_PROTO_SUMMARY.labels(
                    grpc_request=request_name, model_id=model_id
                ).time():
                    if caikit_rpc.output_streaming:
                        response_proto = build_proto_stream(response)
                    else:
                        response_proto = build_proto_response(response)
                return response_proto

    def predict_model(
        self,
        request_name: str,
        model_id: str,
        inference_func_name: str = "run",
        aborter: Optional[RpcAborter] = None,
        **kwargs,
    ) -> Union[DataBase, Iterable[DataBase]]:
        """Run a prediction against the given model using the raw arguments to
        the model's run function.

        Args:
            request_name (str):
                The name of the request message to validate the model's task
            model_id (str):
                The ID of the loaded model
            inference_func_name (str):
                The name of the inference function to run
            aborter (Optional[RpcAborter]):
                If using abortable calls, this is the aborter to use
            **kwargs: Keyword arguments to pass to the model's run function
        Returns:
            response (Union[DataBase, Iterable[DataBase]]):
                The object (unary) or objects (output stream) produced by the
                inference request
        """

        with self._handle_predict_exceptions(model_id, request_name):
            model = self._model_manager.retrieve_model(model_id)
            self._verify_model_task(model)

            # NB: we previously recorded the size of the request, and timed this module to
            # provide a rudimentary throughput metric of size / time
            with alog.ContextLog(
                log.debug,
                "GlobalPredictServicer.Predict.caikit_library_run:%s",
                request_name,
            ):
                model_run_fn = getattr(model, inference_func_name)
                with PREDICT_CAIKIT_LIBRARY_SUMMARY.labels(
                    grpc_request=request_name, model_id=model_id
                ).time():
                    if aborter is not None:
                        work = AbortableAction(aborter, model_run_fn, **kwargs)
                        response = work.do()
                    else:
                        response = model_run_fn(**kwargs)

            # Update Prometheus metrics
            PREDICT_RPC_COUNTER.labels(
                grpc_request=request_name, code=StatusCode.OK.name, model_id=model_id
            ).inc()
            if get_config().runtime.metering.enabled:
                self.rpc_meter.update_metrics(str(type(model)))
            return response

    def stop_metering(self):
        if self._started_metering:
            self.rpc_meter.flush_metrics()
            self.rpc_meter.end_writer_thread()
            self._started_metering = False

    ## Implementation Details ##################################################

    @contextmanager
    def _handle_predict_exceptions(self, model_id: str, request_name: str):
        try:
            yield

        except CaikitRuntimeException as e:
            log_dict = {
                "log_code": "<RUN50530380W>",
                "message": e.message,
                "model_id": model_id,
                "error_id": e.id,
            }
            log.warning({**log_dict, **e.metadata})
            PREDICT_RPC_COUNTER.labels(
                grpc_request=request_name, code=e.status_code.name, model_id=model_id
            ).inc()
            raise e

        # Duplicate code in global_train_servicer
        # pylint: disable=duplicate-code
        except (TypeError, ValueError) as e:
            log_dict = {
                "log_code": "<RUN490439039W>",
                "message": repr(e),
                "model_id": model_id,
                "stack_trace": traceback.format_exc(),
            }
            log.warning(log_dict)
            PREDICT_RPC_COUNTER.labels(
                grpc_request=request_name,
                code=StatusCode.INVALID_ARGUMENT.name,
                model_id=model_id,
            ).inc()
            raise CaikitRuntimeException(
                StatusCode.INVALID_ARGUMENT,
                f"Exception raised during inference. This may be a problem with your input: {e}",
            ) from e

        except Exception as e:
            log_dict = {
                "log_code": "<RUN49049070W>",
                "message": repr(e),
                "model_id": model_id,
                "stack_trace": traceback.format_exc(),
            }
            log.warning(log_dict)
            PREDICT_RPC_COUNTER.labels(
                grpc_request=request_name,
                code=StatusCode.INTERNAL.name,
                model_id=model_id,
            ).inc()
            raise CaikitRuntimeException(
                StatusCode.INTERNAL, "Unhandled exception during prediction"
            ) from e

    def _verify_model_task(self, model: ModuleBase):
        """Raise if the model is not supported for the task"""
        rpc_set: Set[TaskPredictRPC] = set(self._inference_service.caikit_rpcs.values())
        module_rpc: TaskPredictRPC = next(
            (rpc for rpc in rpc_set if model.TASK_CLASS == rpc.task),
            None,
        )

        if not module_rpc:
            raise CaikitRuntimeException(
                status_code=StatusCode.INVALID_ARGUMENT,
                message=f"Inference for model class {type(model)} not supported by this runtime",
            )

    def _build_caikit_library_request_stream(
        self,
        request_stream: Iterable[ProtobufMessage],
        module_signature: CaikitMethodSignature,
        caikit_rpc: TaskPredictRPC,
    ) -> Dict[str, Any]:
        """Builds the kwargs dict to pass to a caikit module.
        Specifically handles the case of constructing input `DataStreams` for some parameters
        which are meant to be streamed in.

        See caikit.runtime.build_caikit_library_request_dict
        """

        def call_build_request_dict(request: ProtobufMessage) -> Dict[str, Any]:
            """This is instead of using a lambda to map each request in the stream"""
            return build_caikit_library_request_dict(request, module_signature)

        streaming_params = caikit_rpc.task.get_required_parameters(input_streaming=True)

        # We need n+1 streams because the first stream is peeked in order to read all the
        # non-streaming parameters off of the first message
        num_streams = 1 + len(streaming_params)
        all_the_streams = itertools.tee(request_stream, num_streams)

        # Read the non-streaming parameters off of the first message in the stream
        stream_num = 0
        kwargs_dict = build_caikit_library_request_dict(
            next(all_the_streams[stream_num]), module_signature
        )
        stream_num += 1

        for param in streaming_params.keys():
            # For each "streaming" parameter, grab one of the tee'd streams and map it to return
            # a `DataStream` of that individual parameter

            def build_getter_from_request_dict(param_name: str) -> Any:
                # This builder is required to correctly closure the `param_name` of the streaming
                # parameter that we're interested in
                def get_fn(request_dict):
                    # Return this parameter out of the request dict
                    return request_dict.get(param_name)

                return get_fn

            param_stream = (
                DataStream.from_iterable(all_the_streams[stream_num])
                .map(call_build_request_dict)
                .map(build_getter_from_request_dict(param_name=param))
            )
            # Add the datastream of this one parameter into the final kwargs dict
            kwargs_dict[param] = param_stream
            stream_num += 1

        return kwargs_dict



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Have pylint ignore Class XXXX has no YYYY member so that we can use gRPC enums.
# pylint: disable=E1101

# Third Party
import grpc

# First Party
import alog

# Local
from caikit.core import MODEL_MANAGER
from caikit.core.data_model import TrainingStatus
from caikit.core.exceptions.caikit_core_exception import (
    CaikitCoreException,
    CaikitCoreStatusCode,
)
from caikit.interfaces.runtime.data_model import (
    TrainingInfoRequest,
    TrainingStatusResponse,
)
from caikit.runtime.types.caikit_runtime_exception import CaikitRuntimeException
from caikit.runtime.utils.servicer_util import raise_caikit_runtime_exception

log = alog.use_channel("MR-SERVICR-I")


class TrainingManagementServicerImpl:
    """This class contains the implementation of all of the RPCs that are required to run a
    service in Model Mesh as a Model-Runtime."""

    def GetTrainingStatus(self, request, context):  # pylint: disable=unused-argument
        """Get the status of a training by ID"""
        training_info_request = TrainingInfoRequest.from_proto(request)
        model_future = self._get_model_future(
            training_info_request.training_id, operation="get_status"
        )
        try:
            reasons = []
            training_info = model_future.get_info()
            if training_info.errors:
                reasons = [str(error) for error in training_info.errors]

            return TrainingStatusResponse(
                training_id=training_info_request.training_id,
                state=training_info.status,
                reasons=reasons,
                submission_timestamp=training_info.submission_time,
                completion_timestamp=training_info.completion_time,
            ).to_proto()
        except CaikitCoreException as err:
            raise_caikit_runtime_exception(exception=err)
        except Exception as err:
            raise CaikitRuntimeException(
                grpc.StatusCode.INTERNAL,
                "Failed to get status for training id {}".format(
                    training_info_request.training_id,
                ),
            ) from err

    def CancelTraining(self, request, context):  # pylint: disable=unused-argument
        """Cancel a training future."""
        training_info_request = TrainingInfoRequest.from_proto(request)
        model_future = self._get_model_future(
            training_info_request.training_id, operation="cancel"
        )
        try:
            model_future.cancel()
            training_info = model_future.get_info()

            reasons = []
            if training_info.errors:
                reasons = [str(error) for error in training_info.errors]

            return TrainingStatusResponse(
                training_id=model_future.id,
                state=training_info.status,
                reasons=reasons,
            ).to_proto()
        except CaikitCoreException as err:
            # In the case that we get a `NOT_FOUND`, we assume that the training was canceled.
            # This is to handle stateful trainers that implement `cancel` by fully deleting
            # the training. NB: Future `GetTrainingStatus` calls for these canceled trainings
            # would raise a not found error to the user.
            if err.status_code == CaikitCoreStatusCode.NOT_FOUND:
                return TrainingStatusResponse(
                    training_id=training_info_request.training_id,
                    state=TrainingStatus.CANCELED,
                ).to_proto()
            raise_caikit_runtime_exception(exception=err)
        except Exception as err:
            log.debug2(
                "Unexpected error trying to cancel training id %s: [%s]",
                training_info_request.training_id,
                err,
            )
            raise CaikitRuntimeException(
                grpc.StatusCode.INTERNAL,
                "Failed to cancel training id {}".format(
                    training_info_request.training_id,
                ),
            ) from err

    @staticmethod
    def _get_model_future(training_id: str, operation: str):
        """Returns a model future, or raises 404 caikit runtime exception on error.
        Wrapped here so that we only catch errors directly in the `trainer.get_model_future` call
        """
        try:
            return MODEL_MANAGER.get_model_future(training_id=training_id)
        except CaikitCoreException as err:
            raise_caikit_runtime_exception(exception=err)
        except Exception as err:
            log.debug2(
                "Caught unexpected exception while trying to look up model future for id %s: [%s]",
                training_id,
                err,
            )
            raise CaikitRuntimeException(
                grpc.StatusCode.INTERNAL,
                "Unexpected error with training id {}. Could not perform {}".format(
                    training_id, operation
                ),
            ) from err



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# Standard
import random

# Third Party
import pytest

# Local
from caikit.core.augmentors.base import AugmentorBase

# Unit Test Infrastructure
from tests.base import TestCaseBase


class StubAugmentor(AugmentorBase):
    """Test class for augmentor core."""

    augmentor_type = str
    # Indicates that this class should not be picked up by augmentor discovery utilities.
    is_test_augmentor = True

    def __init__(self):
        super().__init__(random_seed=1001)

    def _augment(self, obj):
        """Takes a string and returns 'happy', unless the string 'sad' is provided, in which case
        we return an incorrect type.
        """
        if obj == "sad":
            return 1
        else:
            return "happy"


class TestAugmentorBase(TestCaseBase):
    def setUp(self):
        self.test_augmentor = StubAugmentor()

    def test_augment_input_type_enforcement(self):
        """Ensure that input augmentor type is enforced."""
        with self.assertRaises(TypeError):
            self.test_augmentor.augment(True)

    def test_augment_output_type_enforcement(self):
        """Ensure that output augmentor type is enforced."""
        with self.assertRaises(TypeError):
            self.test_augmentor.augment("sad")

    def test_augment_happy_path(self):
        """Ensure that if we pass a happy string, we get a string back."""
        aug_out = self.test_augmentor.augment("happy")
        self.assertIsInstance(aug_out, str)
        self.assertEqual(aug_out, "happy")

    def test_reset_rewinds_random_state(self):
        """Test that if we reset the augmentor, our random state is rewound."""
        # Generate the first 6 numbers using the internal random state
        first_three = [random.random() for x in range(3)]
        next_three = [random.random() for x in range(3)]
        # Make sure the first 3 don't match the second 3
        self.assertNotEqual(first_three, next_three)
        # Then reset and ensure we can regenerate the first 3
        self.test_augmentor.reset()
        reset_three = [random.random() for x in range(3)]
        self.assertEqual(first_three, reset_three)
        # Sanity check to make sure we aren't just getting the same thing both times
        self.assertNotEqual(first_three, next_three)



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Standard
from pathlib import Path
from typing import Dict
import os

# Third Party
import grpc

# First Party
import alog

# Local
from caikit import get_config
from caikit.runtime.types.caikit_runtime_exception import CaikitRuntimeException

log = alog.use_channel("MODEL-SIZER")


class ModelSizer:
    """Model Loader class. The singleton class contains the core implementation details
    for loading models in from S3."""

    __instance = None

    def __init__(self):
        # Re-instantiating this is a programming error
        assert self.__class__.__instance is None, "This class is a singleton!"
        ModelSizer.__instance = self

        # Cache of archive sizes: cos model path -> archive size in bytes
        self._model_archive_sizes: Dict[str, int] = {}

    def get_model_size(self, model_id, local_model_path, model_type) -> int:
        """
        Returns the estimated memory footprint of a model
        Args:
            model_id: The model identifier, used for informative logging
            cos_model_path: The path to the model archive in S3 storage
            model_type: The type of model, used to adjust the memory estimate
        Returns:
            The estimated size in bytes of memory that would be used by loading this model
        """
        # Cache model's size
        if local_model_path not in self._model_archive_sizes:
            self._model_archive_sizes[local_model_path] = self.__get_archive_size(
                model_id, local_model_path
            )

        return self.__estimate_with_multiplier(
            model_id, model_type, self._model_archive_sizes[local_model_path]
        )

    def __estimate_with_multiplier(self, model_id, model_type, archive_size) -> int:
        if (
            model_type
            in get_config().inference_plugin.model_mesh.model_size_multipliers
        ):
            multiplier = (
                get_config().inference_plugin.model_mesh.model_size_multipliers[
                    model_type
                ]
            )
            log.debug(
                "Using size multiplier '%f' for model '%s' to estimate model size",
                multiplier,
                model_id,
            )
        else:
            multiplier = (
                get_config().inference_plugin.model_mesh.default_model_size_multiplier
            )
            log.info(
                "<RUN62161564I>",
                "No configured model size multiplier found for model type '%s' for model '%s'. "
                "Using default multiplier '%f'",
                model_type,
                model_id,
                multiplier,
            )
        return int(archive_size * multiplier)

    def __get_archive_size(self, model_id, local_model_path) -> int:
        try:
            if os.path.isdir(local_model_path):
                # Walk the directory to size all files
                return sum(
                    file.stat().st_size
                    for file in Path(local_model_path).rglob("*")
                    if file.is_file()
                )

            # Probably just an archive file
            return os.path.getsize(local_model_path)
        except FileNotFoundError as ex:
            message = "Failed to estimate size of model '%s', file '%s' not found" % (
                model_id,
                local_model_path,
            )
            log.error("<RUN62168924E>", message)
            raise CaikitRuntimeException(grpc.StatusCode.NOT_FOUND, message) from ex

    @classmethod
    def get_instance(cls) -> "ModelSizer":
        """This method returns the instance of Model Manager"""
        if not cls.__instance:
            cls.__instance = ModelSizer()
        return cls.__instance



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Tests for conversion between python dict and protobuf Struct
"""

# Third Party
from google.protobuf import struct_pb2
import pytest

# Local
from caikit.core.data_model.json_dict import dict_to_struct, struct_to_dict


def test_dict_to_struct_to_dict():
    """Make sure dict_to_struct can handle all variants"""
    raw_dict = {
        "int_val": 1,
        "float_val": 0.42,
        "str_val": "asdf",
        "bool_val": False,
        "null_val": None,
        "list_val": [2, 3.14, "qwer", True, None, [1, 2, 3], {"nested": "val"}],
        "dict_val": {"yep": "works"},
    }

    # Make sure the dict round trips correctly
    struct = dict_to_struct(raw_dict)
    round_trip = struct_to_dict(struct)
    assert round_trip == raw_dict

    # Make sure the struct representation looks right
    assert set(struct.fields) == set(raw_dict)
    assert all(
        getattr(struct.fields[key], struct.fields[key].WhichOneof("kind")) == val
        for key, val in raw_dict.items()
        if not isinstance(val, (list, dict, type(None)))
    )
    assert struct.fields["null_val"].WhichOneof("kind") == "null_value"
    assert struct.fields["null_val"].null_value == struct_pb2.NullValue.NULL_VALUE
    assert isinstance(struct.fields["list_val"].list_value, struct_pb2.ListValue)
    assert len(struct.fields["list_val"].list_value.values) == len(raw_dict["list_val"])
    assert isinstance(struct.fields["dict_val"].struct_value, struct_pb2.Struct)
    assert len(struct.fields["dict_val"].struct_value.fields) == len(
        raw_dict["dict_val"]
    )


def test_dict_to_struct_invalid_value():
    """Make sure that a ValueError is raised if a bad type is encountered"""
    with pytest.raises(ValueError):
        dict_to_struct({"foo": 1, "bar": {"baz": b"asdf"}})



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module holds common utilities for managing arbitrary JSON serializable
dicts as protobuf Struct objects
"""
# Standard
from typing import Dict, List, Optional, Union

# Third Party
from google.protobuf import descriptor, message_factory, struct_pb2

# Type hints for JSON serializable dicts
JsonDictValue = Union[
    int,
    float,
    str,
    bool,
    type(None),
    List["JsonDictValue"],
    "JsonDict",
]
JsonDict = Dict[str, JsonDictValue]

STRUCT_PROTOBUF_NAME = "google.protobuf.Struct"


def dict_to_struct(
    dictionary: JsonDict,
    struct_class: Optional[type] = None,
    value_class: Optional[type] = None,
    list_value_class: Optional[type] = None,
) -> struct_pb2.Struct:
    """Convert a python dict to a protobuf Struct"""
    if struct_class is None:
        struct_class = struct_pb2.Struct
        value_class = struct_pb2.Value
        list_value_class = struct_pb2.ListValue
    else:
        if value_class is None:
            value_class = _get_message_class(
                struct_class.DESCRIPTOR.file.pool.FindMessageTypeByName(
                    "google.protobuf.Value"
                )
            )
        if list_value_class is None:
            list_value_class = _get_message_class(
                struct_class.DESCRIPTOR.file.pool.FindMessageTypeByName(
                    "google.protobuf.ListValue"
                )
            )

    return struct_class(
        fields={
            key: _value_to_struct_value(
                value,
                struct_class=struct_class,
                value_class=value_class,
                list_value_class=list_value_class,
            )
            for key, value in dictionary.items()
        }
    )


def struct_to_dict(struct: struct_pb2.Struct) -> JsonDict:
    """Convert a struct into the equivalent json dict"""
    return {key: _struct_value_to_py(val) for key, val in struct.fields.items()}


## Implementation Details ######################################################


def _value_to_struct_value(value, struct_class, value_class, list_value_class):
    """Recursive helper to convert python values to struct values"""
    if value is None:
        struct_value = value_class(null_value=struct_pb2.NullValue.NULL_VALUE)
    elif isinstance(value, dict):
        struct_value = value_class(
            struct_value=dict_to_struct(
                value,
                struct_class=struct_class,
                value_class=value_class,
                list_value_class=list_value_class,
            )
        )
    elif isinstance(value, list):
        struct_value = value_class(
            list_value=list_value_class(
                values=(
                    _value_to_struct_value(
                        item,
                        struct_class=struct_class,
                        value_class=value_class,
                        list_value_class=list_value_class,
                    )
                    for item in value
                )
            )
        )
    elif isinstance(value, bool):
        struct_value = value_class(bool_value=value)
    elif isinstance(value, int):
        struct_value = value_class(number_value=value)
    elif isinstance(value, float):
        struct_value = value_class(number_value=value)
    elif isinstance(value, str):
        struct_value = value_class(string_value=value)
    else:
        raise ValueError(f"Unsupported value type: {type(value)}")
    return struct_value


def _struct_value_to_py(struct_value: struct_pb2.Value) -> JsonDictValue:
    """Recursive helper to convert struct values to python values"""
    which = struct_value.WhichOneof("kind")
    if which in [None, "null_value"]:
        return None
    if which == "number_value":
        val = struct_value.number_value
        if int(val) == val:
            return int(val)
        return val
    if which in ["string_value", "bool_value"]:
        return getattr(struct_value, which)
    if which == "struct_value":
        return struct_to_dict(struct_value.struct_value)
    if which == "list_value":
        return [_struct_value_to_py(item) for item in struct_value.list_value.values]


def _get_message_class(
    desc: descriptor.Descriptor,
) -> message_factory.message.Message:
    """Helper to get the concrete protobuf class from a descriptor. This
    supports compatibility between protobuf 3.X and 4.X
    """
    if hasattr(message_factory, "GetMessageClass"):
        return message_factory.GetMessageClass(desc)  # pragma: no cover
    return message_factory.MessageFactory().GetPrototype(desc)  # pragma: no cover
